from .image import Image
from .db_images import DbImages
from typing import List
from .fieldnames import *
from ..models.utils import load_model
from ..datasets.utils import load_binary_image_classification_dataset
from torch.utils.data import DataLoader
from datetime import datetime as dt
import torch
from ..db.conversions import binary_classification_pred_to_db_value
from tqdm.auto import tqdm
from pathlib import Path


class AiPredictor:
    """[summary]
    """

    def __init__(self,
        ai_name,
        cfg,
        db_images: DbImages,
    ):
        self.ai_name = ai_name
        self.specs = cfg["models"][ai_name]
        self.cfg = cfg
        self.version = self.specs[FIELDNAME_AI_VERSION]
        self.image_scaling = self.specs[FIELDNAME_IMAGE_SCALING]
        self.base_paths_models = Path(cfg["base_path_models"])

    def load_model(self, specs = None, _eval: bool = True):
        self.model = None
        if not specs:
            specs = self.specs
        
        print("Loading in Prediction Mode!")
        self.model = load_model(self.ai_name, specs[FIELDNAME_AI_VERSION], _eval, self.base_paths_models)
        if self.model:
            return True

    def load_dataset(self, image_list: List[dict]):
        self.dataset = None
        paths = [_[FIELDNAME_IMAGE_PATH] for _ in image_list]
        ids = [str(_["_id"]) for _ in image_list]

        self.dataset = load_binary_image_classification_dataset(paths, ids, self.image_scaling, training = False)

        if self.dataset:
            return True

    def load_dataloader(
            self,
            dataset = None,
            batch_size: int = 100,
            num_workers = 0 ################ Change?
        ):
        if not dataset:
            dataset = self.dataset

        self.dataloader = None
        self.dataloader = DataLoader(
            dataset, batch_size=batch_size, shuffle=False, sampler=None,
            batch_sampler=None, num_workers=num_workers, collate_fn=None,
            pin_memory=False, drop_last=True, timeout=0,
            worker_init_fn=None,prefetch_factor=2,
            persistent_workers=False
        )

        if self.dataloader:
            return True


    def predict_images(self, targets: List[str] = None, model = None, dataloader = None, specs = None):
        if not dataloader:
            dataloader = self.dataloader

        if not model:
            model = self.model

        if not specs:
            specs = self.specs

        if not targets:
            targets = [self.ai_name]
        
        creation_date = dt.now()
        print(targets)
        predicted_batches = []
        for batch_images, batch_labels in tqdm(dataloader):
            if torch.cuda.is_available():
                _pred = model(batch_images.to(0))#.to(0))
            else:
                _pred = model(batch_images)
            
            update = binary_classification_pred_to_db_value(
                _pred,
                batch_labels,
                version = specs[FIELDNAME_AI_VERSION],
                targets = targets,
                creation_date=creation_date)
            
            predicted_batches.append(update)
        
        return targets, predicted_batches

        


import numpy as np
import time

from numpy.random import default_rng
from numba import jit
from collections import namedtuple

RANDOM_SEED = 23
Solution = namedtuple("Solution", "cost index")


@jit(nopython=True)
def index_to_str(idx):
    """
    Generates a string representation from an index array.
    @param idx: The NumPy boolean index array.
    @return: The string representation of the array.
    """
    num_chars = int(idx.shape[0] / 6 + 0.5)
    s = ""
    for i in range(num_chars):
        b = i * 6
        six = idx[b:b+6]
        c = 0
        for j in range(six.shape[0]):
            c = c * 2 + int(six[j])
        s = s + chr(c+32)
    return s


def create_initial_solution(sample_counts, p):
    """
    Creates the initial solution using a random shuffle
    @param sample_counts: The problem array.
    @param p: The validation set proportion.
    @return: A solution array.
    """
    rng = default_rng(seed=RANDOM_SEED)
    group_count = sample_counts.shape[0]
    idx = np.zeros(group_count, dtype=bool)
    rnd_idx = rng.permutation(group_count)
    start_count = 0
    sample_size = sample_counts[:, 0].sum()
    for i in range(group_count):
        start_count += sample_counts[rnd_idx[i], 0]
        idx[rnd_idx[i]] = True
        if start_count > sample_size * p:
            break
    return idx


@jit(nopython=True)
def calculate_cost(sample_counts, idx, p):
    """
    Calculates the cost of a given solution
    @param sample_counts: The problem array.
    @param idx: The solution array to evaluate.
    @param p: The train/validation split proportion.
    @return: The cost value.
    """
    train_count = sample_counts[~idx, 0].sum()
    valid_count = sample_counts[idx, 0].sum()
    total_count = train_count + valid_count

    cost = (valid_count - total_count * p) ** 2

    for i in range(1, sample_counts.shape[1]):
        r = sample_counts[:, i].sum() / total_count
        cost += (sample_counts[idx, i].sum() - r * sample_counts[idx, 0].sum()) ** 2
    return cost / 2.0


def calculate_cost_grad(sample_counts, idx, p):
    """
    Calculates the cost gradient of a given solution
    @param sample_counts: The problem array.
    @param idx: The solution array to evaluate.
    @param p: The train/validation split proportion.
    @return: The cost value.
    """
    grad = np.zeros(sample_counts.shape[1])

    total_count = sample_counts[:, 0].sum()
    valid_count = sample_counts[idx, 0].sum()

    grad[0] = total_count * p - valid_count

    for i in range(1, sample_counts.shape[1]):
        r = sample_counts[:, i].sum() / total_count
        grad[i] = r * sample_counts[idx, 0].sum() - sample_counts[idx, i].sum()
    return grad


def cosine_similarity(sample_counts, idx, cost_grad):
    """
    Calculates the cosine similarity vector between the problem array
    and the cost gradient vector
    @param sample_counts: The problem array.
    @param idx: The solution vector.
    @param cost_grad: The cost gradient vector.
    @return: The cosine similarity vector.
    """
    c = np.copy(sample_counts)
    c[idx] = -c[idx]            # Reverse direction of validation vectors
    a = np.inner(c, cost_grad)
    b = np.multiply(np.linalg.norm(c, axis=1), np.linalg.norm(cost_grad))
    return np.divide(a, b)


def euclidean_similarity(sample_counts, idx, cost_grad):
    c = np.copy(sample_counts)
    c[idx] = -c[idx]
    return np.linalg.norm(c - cost_grad, axis=1)


def generate_cosine_move(sample_counts, idx, p, expanded_set, intensify):
    """
    Generates a new move using the cosine similarity.
    @param sample_counts: The problem array.
    @param idx: The solution vector.
    @param p: The validation set proportion.
    @param expanded_set: The set of expanded solutions.
    @param intensify: Intensification / diversification flag.
    @return: The new solution vector.
    """
    cost_grad = calculate_cost_grad(sample_counts, idx, p)
    similarity = cosine_similarity(sample_counts, idx, cost_grad)
    sorted_ixs = np.argsort(similarity)
    if intensify:
        sorted_ixs = np.flip(sorted_ixs)
    for i in sorted_ixs:
        move = np.copy(idx)
        move[i] = not move[i]
        if index_to_str(move) not in expanded_set:
            return move
    return None


def generate_euclidean_move(sample_counts, idx, p, expanded_set, get_min):
    cost_grad = calculate_cost_grad(sample_counts, idx, p)
    similarity = euclidean_similarity(sample_counts, idx, cost_grad)
    sorted_ixs = np.argsort(similarity)
    if not get_min:
        sorted_ixs = np.flip(sorted_ixs)
    for i in sorted_ixs:
        move = np.copy(idx)
        move[i] = not move[i]
        if index_to_str(move) not in expanded_set:
            return move
    return None


def generate_moves(idx, expanded_set):
    """
    Generator for all acceptable moves from a previous solution.
    @param idx: The solution vector.
    @param expanded_set: The set of expanded solutions.
    """
    for i in range(idx.shape[0]):
        move = np.copy(idx)
        move[i] = not move[i]
        if index_to_str(move) not in expanded_set:
            yield move


def generate_counts(num_groups, num_classes,
                    min_group_size, max_group_size,
                    max_group_percent):
    """
    Generates a problem matrix from the given parameters.
    @param num_groups: The number of data groups.
    @param num_classes: The number of classes.
    @param min_group_size: The minimum group size.
    @param max_group_size: The maximum group size.
    @param max_group_percent: The maximum class percent.
    @return: The problem matrix.
    """
    rng = default_rng(seed=RANDOM_SEED)
    sample_cnt = np.zeros((num_groups, num_classes), dtype=int)
    sample_cnt[:, 0] = rng.integers(low=min_group_size, high=max_group_size, size=num_groups)

    for i in range(1, num_groups):
        for j in range(1, num_classes):
            sample_cnt[i, j] = rng.integers(low=0, high=max_group_percent * sample_cnt[i, 0] - sample_cnt[i, 1:j+1].sum())
    return sample_cnt


class BaseSolver(object):

    def __init__(self, problem, candidate, p):
        self.problem = problem
        self.p = p
        self.incumbent = Solution(calculate_cost(problem, candidate, p), candidate)


class SearchSolver(BaseSolver):

    def __init__(self, problem, candidate, p):
        super(SearchSolver, self).__init__(problem, candidate, p)

    def solve(self, min_cost, max_empty_iterations,
              max_intensity_iterations, verbose=True):
        """
        Uses the search solver to calculate the best split.
        @param min_cost: Minimum cost criterion.
        @param max_empty_iterations: Maximum number of non-improving iterations.
        @param max_intensity_iterations: Maximum number of intensity iterations.
        @param verbose: Verbose flag.
        @return: The incumbent solution.
        """
        terminated = False
        intensify = True
        expanded_set = set()
        solution = self.incumbent
        n = 0
        n_intensity = 0

        while not terminated:
            move_list = generate_moves(solution.index, expanded_set)
            cost_list = [Solution(calculate_cost(self.problem, move, self.p), move)
                         for move in move_list]
            cost_list.sort(key=lambda t: t[0], reverse=not intensify)
            intensify = True

            if len(cost_list) > 0:
                solution = cost_list[0]
                expanded_set.add(index_to_str(solution.index))
                if solution.cost < self.incumbent.cost:
                    self.incumbent = solution
                    n = 0
                    n_intensity = 0
                    intensify = True
                    if verbose:
                        print(self.incumbent.cost)
                else:
                    # Diversify?
                    if n_intensity > max_intensity_iterations:
                        intensify = False
                        n_intensity = 0
            else:
                terminated = True

            n += 1
            n_intensity += 1
            if n > max_empty_iterations or self.incumbent.cost < min_cost:
                terminated = True
        return self.incumbent


class GradientSolver(BaseSolver):

    def __init__(self, problem, candidate, p):
        super(GradientSolver, self).__init__(problem, candidate, p)

    def solve(self, min_cost, max_empty_iterations,
              max_intensity_iterations, verbose=True):
        """
        Uses the gradient solver to calculate the best split.
        @param min_cost: Minimum cost criterion.
        @param max_empty_iterations: Maximum number of non-improving iterations.
        @param max_intensity_iterations: Maximum number of intensity iterations.
        @param verbose: Verbose flag.
        @return: The incumbent solution.
        """
        terminated = False
        intensify = True
        expanded_set = set()
        solution = self.incumbent
        n = 0
        n_intensity = 0

        while not terminated:
            move = generate_cosine_move(self.problem, solution.index, self.p,
                                        expanded_set, intensify)
            intensify = True

            if move is not None:
                solution = Solution(calculate_cost(self.problem, move, self.p), move)
                expanded_set.add(index_to_str(solution.index))
                if solution.cost < self.incumbent.cost:
                    self.incumbent = solution
                    n = 0
                    n_intensity = 0

                    if verbose:
                        print(self.incumbent.cost)
                else:
                    if n_intensity > max_intensity_iterations:
                        intensify = False
                        n_intensity = 0
            else:
                terminated = True

            n += 1
            n_intensity += 1
            if n > max_empty_iterations or self.incumbent.cost < min_cost:
                terminated = True

        return self.incumbent


def print_solution(problem, solution, p):
    idx = solution.index
    valid_count = problem[idx, 0].sum()
    train_count = problem[~idx, 0].sum()
    total_count = train_count + valid_count

    print(solution.cost)
    print(p, valid_count / total_count)
    for i in range(1, problem.shape[1]):
        r = problem[:, i].sum() / total_count
        print(r, problem[idx, i].sum() / problem[idx, 0].sum())


def main():
    num_groups = 250            # Number of groups to simulate
    num_classes = 2             # Number of classes
    max_group_size = 10000      # Maximum group size
    max_group_perc = 0.4        # Maximum proportion for each class
    p = 0.3                     # Validation split proportion

    max_empty_iterations = 100
    max_intensity_iterations = 10
    min_cost = 10000

    sample_cnt = generate_counts(num_groups, num_classes,
                                 min_group_size=10,
                                 max_group_size=max_group_size,
                                 max_group_percent=max_group_perc)

    solution_arr = create_initial_solution(sample_cnt, p)

    s_solver = SearchSolver(sample_cnt, solution_arr, p)
    g_solver = GradientSolver(sample_cnt, solution_arr, p)

    start = time.time()
    solution = s_solver.solve(min_cost, max_empty_iterations, max_intensity_iterations, verbose=False)
    print(time.time() - start)

    print_solution(sample_cnt, solution, p)
    print()

    start = time.time()
    solution = g_solver.solve(min_cost, max_empty_iterations, max_intensity_iterations, verbose=False)
    print(time.time() - start)

    print_solution(sample_cnt, solution, p)


# if __name__ == "__main__":
#     main()


from bson import ObjectId
from typing import List
from .get_objects import get_images_by_id_list, get_intervention
from tqdm import tqdm

def video_keys_unique(db_interventions):
    video_keys = db_interventions.find({"video_key": {"$exists": True}}, {"video_key": 1})
    video_keys = [_ for _ in video_keys]
    _keys = []
    _duplicates = []

    for _ in video_keys:
        if _["video_key"] in _keys:
            _duplicates.append(_["video_key"])
        else:
            _keys.append(_["video_key"])
        
    duplicates = [_ for _ in video_keys if _["video_key"] in _duplicates]

    return duplicates
    
def ids_extern_unique(db_interventions):
    ids_extern = db_interventions.find({"ids_extern": {"$exists": True}}, {"ids_extern": 1})
    ids_extern = [_ for _ in ids_extern]
    _keys = []
    _duplicates = []

    for _ in ids_extern:
        if _["ids_extern"] in _keys:
            _duplicates.append(_["ids_extern"])
        else:
            _keys.append(_["ids_extern"])
        
    duplicates = [_ for _ in ids_extern if _["ids_extern"] in _duplicates]

    return duplicates

def check_intervention_frames_extracted(video_key, db_images, db_interventions):
    intervention = get_intervention(video_key, db_interventions)
    id_list = [_id for n, _id in intervention.frames.items()]
    images = get_images_by_id_list(id_list, db_images)

    is_extracted_dict = {}
    for image in images:
        if image.metadata.path:
            is_extracted_dict[image.metadata.frame_number] = image.metadata.path.exists()
        else: 
            is_extracted_dict[image.metadata.frame_number] = False
    return is_extracted_dict


from .base import PyObjectId
from bson import ObjectId
from pydantic import BaseModel, Field
from typing import List


class DbTestSet(BaseModel):
    id: PyObjectId = Field(alias="_id")
    name: str
    ids: List[PyObjectId]
    
    class Config:
        allow_population_by_field_name = True
        json_encoders = {ObjectId: str}
        schema_extra = {
            "example": {}
        }



from .db import Db

class AiTrainer:
    """[summary]
    """
    
    def __init__(self, ai_name: str, db: Db):
        self.ai_name = ai_name
        self.db = db
        self.db_images = db.db_images
        self.db_interventions = db.db_interventions
        self.cfg = db.cfg
        self.specs = self.cfg["models"][ai_name]


    def train_ai(self):
        # orig_cwd = hydra.utils.get_original_cwd()
        # configuration for folder structure during model generation in: hydra/default.yaml
        pass



from ..classes.train_data import TrainData
from ..fieldnames import *


def get_train_data_query(name, exclude_intervention_ids, exclude_origins):
    query = {
        "$match": {
            f"{IMG_ANNOTATIONS}.{name}": {"$exists": True},
            IMG_INTERVENTION_ID: {"$nin": exclude_intervention_ids},
            ORIGIN: {"$nin": exclude_origins},
            f"{METADATA}.is_extracted": True
        }
    }

    return query

def image_list_to_train_data(name, image_list, label_type, choices):
    paths = []
    labels = []
    origins = []
    intervention_ids = []
    image_ids = []

    for image in image_list:
        assert choices == image[IMG_ANNOTATIONS][name]["choices"]
        paths.append(image[METADATA]["path"])
        labels.append(image[IMG_ANNOTATIONS][name]["value"])
        origins.append(image[ORIGIN])
        intervention_ids.append(image[IMG_INTERVENTION_ID])
        image_ids.append(image["_id"])

    train_data = {
        "name": name,
        "paths": paths,
        "labels": labels,
        "prediction_type": label_type,
        "choices": choices,
        "origins": origins,
        "intervention_ids": intervention_ids,
        "image_ids": image_ids,
    }

    train_data = TrainData(**train_data)
    return train_data


from ..classes.annotation import MultilabelAnnotation
from collections import defaultdict
from tqdm import tqdm

def get_multiclass_image_dict(label_list, db_images):
    updates = defaultdict(list)
    conflicts = []
    for label in label_list:
        check_labels = [_ for _ in label_list if _ != label]
        cursor = db_images.find({f"annotations.{label}.value": True})
        for image in tqdm(cursor):
            conflict = False
            for _label in check_labels:
                if _label in image["annotations"]:
                    if image["annotations"][_label]["value"] == True:
                        conflict = True
            if conflict:
                conflicts.append(image["_id"])
            else: 
                updates[label].append(image["_id"])

    conflicts = list(set(conflicts))
    return updates, conflicts

def get_outside_images(db_images):
    _label = "body"
    cursor = db_images.find({f"annotations.{_label}.value": False})

    return [_["_id"] for _ in cursor]

def set_compound_labels(
    name,
    base_annotation_dict,
    ai_config,
    db_images,
    exclude_initially = [
            "outside",
            "blurry"
        ]
    ):
    label_list = [_ for _ in ai_config.choices if _ not in exclude_initially]
    updates, conflicts = get_multiclass_image_dict(label_list, db_images)

    if "outside" in ai_config.choices:
        updates["outside"] = get_outside_images(db_images)

    image_updates = {}
    for label, ids in updates.items():
        value = ai_config.choices.index(label)
        for _id in ids:
            image_updates[_id] = value

    if "blurry" in ai_config.choices:
        blurry_image_ids = [_["_id"] for _ in db_images.find({"annotations.blurry.value": True})]

        value = ai_config.choices.index("blurry")
        for _id in blurry_image_ids:
            image_updates[_id] = value

    image_annotations = {}

    for _id, value in image_updates.items():
        _annotation = base_annotation_dict.copy()
        _annotation["value"] = value
        image_annotations[_id] = MultilabelAnnotation(**_annotation)

    for _id, annotation in tqdm(image_annotations.items()):
        db_images.update_one({"_id": _id}, {"$set": {f"annotations.{name}": annotation.dict()}})

    


from pathlib import Path
import requests
from typing import List

from ukw_ml_tools.classes.fieldnames import FIELDNAME_VIDEO_KEY, FIELDNAME_VIDEO_PATH
from ukw_ml_tools.classes.utils import get_intervention_db_template, check_extern_video_annotation_timestamp


class DbExtern:
    """[summary]
    """

    def __init__(self, cfg: dict):
        # self.http_prefix = 
        self.url = cfg["http_prefix"] + cfg["url_endobox_extreme"] + ":" + cfg["port_webserver"] + "/data"
        self.fields_extern = cfg["db_extern_fields"]
        self.user = cfg["user_webserver"]
        self._p = cfg["password_webserver"]
        self._auth = (self.user, self._p)
        self.lookup_extern_fieldtypes = {
            "int": int,
            "str": str,
        }
        self.time_format = cfg["db_extern_date_pattern"]
        self.user_priority = cfg["db_extern_annotation_user_priority"]


    def convert_intervention_external_to_internal(self, intervention):
        valid = self.validate_extern_intervention(intervention)
        if not valid:
            raise Exception

        intervention_internal = get_intervention_db_template(intervention_dict = intervention)

        for field, value in intervention.items():
            if field in self.fields_extern:
                if "map_value" in self.fields_extern[field]:
                    value = self.fields_extern[field]["map_value"][value]
                internal_field_name = self.fields_extern[field]["internal_field"]
                intervention_internal[internal_field_name] = value

        video_key = Path(intervention_internal[FIELDNAME_VIDEO_PATH]).name
        intervention_internal[FIELDNAME_VIDEO_KEY] = video_key

        return intervention_internal


    def get_extern_interventions(self) -> List[dict]:
        r = requests.get(f"{self.url}/GetVideosExtern", auth=self._auth, verify = False)
        assert r.status_code == 200

        return r.json()

    def get_extern_annotations(self):
        r = requests.get(f"{self.url}/GetVideosWithAnnotations", auth = self._auth, verify = False)
        assert r.status_code == 200

        return r.json()

    def get_extern_video_annotation(self, video_key):
        r = requests.get(self.url+"/GetAnnotationsByVideoName/"+video_key, auth = self._auth, verify = False)
        annotations = r.json()

        return annotations

    def extract_frames(self):
        pass

    # Validation
    def validate_field_extern(self, field, value) -> bool:
        if field not in self.fields_extern:
            return True
        
        field_config = self.fields_extern[field]
        target_type = field_config["type"]
        target_type = self.lookup_extern_fieldtypes[target_type]
        
        if not field_config["required"] and value == None:
            return True
        else:    
            _check_value = isinstance(value, target_type)
            return _check_value


    def validate_extern_intervention(self, intervention: dict) -> bool:
        '''Only validates fields mentioned in cfg["db_extern_fields"]'''
        valid = True

        for field, value in intervention.items():
            field_valid = self.validate_field_extern(field, value)
            if not field_valid:
                valid = False
        
        return valid




from a_ukw_ml_tools.mongodb.test_data import get_test_data_ids
from a_ukw_ml_tools.mongodb.train_data import get_train_data_query, image_list_to_train_data
from a_ukw_ml_tools.mongodb.base import filter_images_by_frame_diff

def create_new_train_data(name, ai_config, db_images, db_test_data):

    choices = ai_config.choices
    label_type = ai_config.label_type
    exclude_origins = ai_config.sampler_settings.exclude_origins
    test_sets = ai_config.sampler_settings.test_sets
    exclude_intervention_ids = []
    for test_set in test_sets:
        exclude_intervention_ids.extend(get_test_data_ids(test_set, db_test_data))

    aggregation = [
        get_train_data_query(name, exclude_intervention_ids, exclude_origins),
        # {
        #     "$project": {
        #         "paths": "$metadata.path",
        #         "labels": f"$annotations.{name}.value",
        #         "origins": "$origin",
        #         "intervention_ids": "$intervention_id",
        #         "choices": f"$annotations.{name}.choices"
        #     }
        # }
    ]

    
    cursor = db_images.aggregate(aggregation)
    images = filter_images_by_frame_diff(cursor, ai_config.sampler_settings.min_frame_diff)
    train_data = image_list_to_train_data(name, label_type, choices, images)

    return train_data


def get_intervention_labels(intervention_id, db_images, select = "predictions"):
    """
    Returns one record for each label (_id, name, value, intervention_id, framenumber)
    """
    aggregation = [
        {"$match": {"intervention_id": intervention_id}},
        {
            "$project": {
                select: {"$objectToArray": f"${select}"},
                "video_key": "$video_key",
                "intervention_id": "$intervention_id",
                "frame_number": "$metadata.frame_number"
            }
        },
        {
            "$unwind":  f"${select}"
        },
        {
            "$project": {
                "intervention_id": "$intervention_id",
                "video_key": "$video_key",
                "frame_number": "$frame_number",
                "_type": select,
                "name": f"${select}.v.name",
                "value": f"${select}.v.value",
                "raw": f"${select}.v.raw"
            }
        }
    ]

    r = db_images.aggregate(aggregation)
    records = [_ for _ in r]
    return records





from pymongo.collection import Collection
from bson.objectid import ObjectId
from collections import Counter
import warnings
from pathlib import Path
from typing import List
from pymongo.cursor import Cursor
import os

from tqdm import tqdm

from .terminology import Terminology
from .utils import *
from .fieldnames import *
import shutil


class DbInterventions:
    """Add Documentation
    asd
    """
    def __init__(self, db_interventions: Collection, cfg: dict):
        self.db = db_interventions
        self.cfg = cfg
        self.video_key_re_date_pattern = cfg["video_key_re_date_pattern"]
        self.video_key_dt_date_pattern = cfg["video_key_dt_date_pattern"]
        self.paths = {
            "base_path_frames": Path(cfg["base_path_frames"])
        }
        self.terminology = None

    def get_by_id(self, _id: ObjectId):
        return self.db.find_one({"_id": _id})

    # Terminology
    def init_terminology(self, terminology_path: Path = None, terminology_type: int = 0):
        self.terminology = Terminology(self.cfg, terminology_path, terminology_type)

    def apply_terminology(self, intervention_list: List[dict] = None, set_tokens_in_db: bool = False, return_tokens = True, preprocess_text: bool = True):
        if set_tokens_in_db:
            assert return_tokens
        
        if not self.terminology:
            self.init_terminology()

        terminology_type = self.terminology.terminology_type

        if not intervention_list:
            intervention_list = self.get_interventions_with_report_or_patho()
        
        results = {}
        intervention_list = [_ for _ in intervention_list]

        for intervention in tqdm(intervention_list):
            if FIELDNAME_REPORT_RAW in intervention:
                text = intervention[FIELDNAME_REPORT_RAW]
                if isinstance(text, dict):
                    text = dict_of_lists_to_text(text)
                elif isinstance(text, type(None)):
                    text = ""
                result_report = self.terminology.get_terminology_result(text, terminology_type, return_tokens, preprocess_text)
            else:
                result_report = {}
            if FIELDNAME_PATHO_RAW in intervention:
                text = intervention[FIELDNAME_PATHO_RAW]
                if isinstance(text, dict):
                    text = dict_of_lists_to_text(text)
                elif isinstance(text, type(None)):
                    text = ""
                result_patho = self.terminology.get_terminology_result(text, terminology_type, return_tokens, preprocess_text)
            else:
                result_patho = {}
            result = {FIELDNAME_TOKENS_REPORT: result_report, FIELDNAME_TOKENS_PATHO: result_patho}
            results[intervention["_id"]] = result

            if set_tokens_in_db:
                self.db.update_one({"_id": intervention["_id"]}, {"$set": {FIELDNAME_TOKENS: result}})

        return results


    # Queries
    ## Create
    def create_new_from_external(self, intervention_dict: dict):
        video_key = intervention_dict[FIELDNAME_VIDEO_KEY]
        _ = self.db.find_one(field_value_query(FIELDNAME_VIDEO_KEY, video_key))
        if _:
            return False

        intervention_date = datetime_from_video_key(
            video_key,
            self.video_key_re_date_pattern,
            self.video_key_dt_date_pattern
        )

        metadata = get_video_meta(Path(intervention_dict[FIELDNAME_VIDEO_PATH]))
        intervention_dict[FIELDNAME_VIDEO_METADATA] = metadata

        if intervention_date:
            intervention_dict[FIELDNAME_INTERVENTION_DATE] = intervention_date

    
        result = self.db.insert_one(intervention_dict)
        return result.inserted_id

    def create_new_intervention_from_video_file(self, source_video_path: Path, target_video_dir: Path, origin: str, intervention_type: str):
        intervention_dict = get_intervention_db_template()
        
        assert intervention_type in self.cfg["intervention_types"]
        assert source_video_path.is_file()
        assert target_video_dir.is_dir()

        video_key = source_video_path.name
        target_path = target_video_dir.joinpath(video_key)
        assert not target_path.exists()
        assert not self.db.find_one({FIELDNAME_VIDEO_KEY: video_key})
        intervention_date = datetime_from_video_key(
            video_key,
            self.video_key_re_date_pattern,
            self.video_key_dt_date_pattern
        )

        metadata = get_video_meta(source_video_path)

        intervention_dict[FIELDNAME_VIDEO_PATH] = target_path.as_posix()
        intervention_dict[FIELDNAME_VIDEO_KEY] = video_key
        intervention_dict[FIELDNAME_ORIGIN] = origin
        intervention_dict[FIELDNAME_VIDEO_METADATA] = metadata
        intervention_dict[FIELDNAME_INTERVENTION_TYPE] = intervention_type
        if intervention_date:
            intervention_dict[FIELDNAME_INTERVENTION_DATE] = intervention_date

        shutil.copyfile(source_video_path, target_path)
        result = self.db.insert_one(intervention_dict)

        return result.inserted_id
    
    ## Read
    def get_interventions_with_video(self, as_list: bool=False) -> (Cursor or List):
        interventions = self.db.find(field_exists_query(FIELDNAME_VIDEO_PATH))

        if as_list: return [_ for _ in interventions]
        else: return interventions

    def get_interventions_with_freezes(self, as_list: bool = False):
        interventions = self.db.find(fieldvalue_nin_list_query(FIELDNAME_FREEZES, [{}]))

        if as_list: return [_ for _ in interventions]
        else: return interventions

    def get_interventions_with_frames(self, as_list: bool = False):
        interventions = self.db.find(fieldvalue_nin_list_query(FIELDNAME_FRAMES, [{}]))

        if as_list: return [_ for _ in interventions]
        else: return interventions

    def get_interventions_with_reports(self, as_list: bool = False):
        interventions = self.db.find({FIELDNAME_REPORT_RAW: {"$exists": True}})

        if as_list: return [_ for _ in interventions]
        else: return interventions

    def get_interventions_with_patho(self, as_list: bool = False):
        interventions = self.db.find({FIELDNAME_PATHO_RAW: {"$exists": True}})

        if as_list: return [_ for _ in interventions]
        else: return interventions

    def get_interventions_with_report_or_patho(self, as_list: bool = False):
        aggregation = [
            {
                "$match": {
                    "$or": [
                        {FIELDNAME_PATHO_RAW: {"$exists": True, "$nin": [None, ""]}},
                        {FIELDNAME_REPORT_RAW: {"$exists": True, "$nin": [None, ""]}}
                    ]
                }
            }
        ]

        interventions = self.db.aggregate(aggregation)

        if as_list: return [_ for _ in interventions]
        else: return interventions

    def get_distinct_values(self, field_str: str):
        return self.db.distinct(field_str)

    # Stats
    def calculate_stats(self, return_records: bool = True):
        self.stats_queries = {
            "entity": PREFIX_INTERVENTION,
            "get_count": [
                {f"{PREFIX_EXISTS}.{FIELDNAME_PATHO_RAW}": {FIELDNAME_PATHO_RAW: {"$exists": True}}},
                {f"{PREFIX_EXISTS}.{FIELDNAME_REPORT_RAW}": {FIELDNAME_REPORT_RAW: {"$exists": True}}},
                {f"{PREFIX_EXISTS}.{FIELDNAME_REPORT_RAW},{FIELDNAME_PATHO_RAW}": {
                    FIELDNAME_PATHO_RAW: {"$exists": True},
                    FIELDNAME_REPORT_RAW: {"$exists": True}
                }},
                {f"{PREFIX_EXISTS}.{FIELDNAME_VIDEO_KEY}": {FIELDNAME_VIDEO_KEY: {"$exists": True}}},
            ],
            "get_grouped_count": [
                {
                    "fieldname": FIELDNAME_ORIGIN,
                    "additional_match_conditions": None
                },
                {
                    "fieldname": FIELDNAME_ORIGIN,
                    "additional_match_conditions": {
                        f"{PREFIX_EXISTS}.{FIELDNAME_PATHO_RAW}": {FIELDNAME_PATHO_RAW: {"$exists": True}}
                    }
                },
                {
                    "fieldname": FIELDNAME_ORIGIN,
                    "additional_match_conditions": {
                        f"{PREFIX_EXISTS}.{FIELDNAME_REPORT_RAW}": {FIELDNAME_REPORT_RAW: {"$exists": True}}
                    }
                },
                {
                    "fieldname": FIELDNAME_ORIGIN,
                    "additional_match_conditions": {
                        f"{PREFIX_EXISTS}.{FIELDNAME_REPORT_RAW},{FIELDNAME_PATHO_RAW}": {FIELDNAME_REPORT_RAW: {"$exists": True}, FIELDNAME_PATHO_RAW: {"$exists": True}}
                    }
                },
            ],
            "get_grouped_dict_count": []
        }

        stats = calculate_stats_dict(self.db, self.stats_queries, return_records = return_records)
        return stats

    # Update
    def update_n_extracted_frames(self):
        aggregation = [
            {
                "$project": {
                    "_id": 1,
                    "frame_count": {
                        "$size": {
                            "$objectToArray": f"${FIELDNAME_FRAMES}"
                        }
                    }
                }
            },
            {
                "$match": {
                }
            }
        ]
        interventions = self.db.aggregate(aggregation)
        for _ in interventions:
            self.db.update_one({"_id": _["_id"]}, {"$set": {FIELDNAME_N_EXTRACTED_FRAMES: _["frame_count"]}})
    # Validation
    def validate_video_keys(self) -> List[str]:
        """Filters for non unique video_keys

        Args:
            db_interventions (Collection):

        Returns:
            List: List of non-unique video keys
        """
        interventions = self.db.find(
            {FIELDNAME_VIDEO_KEY: {"$exists": True}}, {FIELDNAME_VIDEO_KEY: 1}
        )
        keys = [_[FIELDNAME_VIDEO_KEY] for _ in interventions]
        duplicates = [key for key, count in Counter(keys).items() if count > 1]

        if duplicates:
            warnings.warn("Non unique video keys detected")
            warnings.warn("\n".join(duplicates))

        return duplicates

    def validate_video_paths(self) -> List[dict]:
        _interventions = self.get_interventions_with_video(as_list = True)

        path_is_none = []
        interventions = []

        for _ in _interventions:
            if not _[FIELDNAME_VIDEO_PATH]:
                path_is_none.append(str(_["_id"]))
            else:
                interventions.append(_)

        not_exists = [
            str(_["_id"]) for _ in interventions if not Path(_[FIELDNAME_VIDEO_PATH]).exists()
        ]

        if not_exists:
            warnings.warn("Videos pointing to non existing file detected")

        if path_is_none:
            warnings.warn("Videos with path == None or path == '' found")

        return not_exists, path_is_none

    def validate_frame_dirs(self, create_if_missing: bool = False) -> List[dict]:
        interventions = self.get_interventions_with_video(as_list = False)
        
        no_frame_dir = [
            str(_["_id"]) for _ in interventions if not generate_frame_path(
                video_key=_[FIELDNAME_VIDEO_KEY],
                base_path_frames=self.paths["base_path_frames"]
            ).exists()
        ]

        if no_frame_dir:
            if create_if_missing:
                for intervention in no_frame_dir:
                    _path = generate_frame_path(
                        video_key=intervention[FIELDNAME_VIDEO_KEY],
                        base_path_frames=self.paths["base_path_frames"]
                    )
                    os.mkdir(_path)
            else:
                warnings.warn("Videos without frame directoy detected")

        return no_frame_dir




from ..mongodb.get_objects import get_intervention
from typing import Tuple, List
from itertools import groupby
from operator import itemgetter
from ..classes.base import Flank
from collections import defaultdict
from ..classes.annotation import VideoSegmentationAnnotation
import numpy as np
from sklearn.preprocessing import label_binarize

def get_consecutive_ranges(list) -> List[Tuple]:
    range_tuples = []
    for k, g in groupby(enumerate(list), lambda x: x[0]-x[1]):
        range_tuples.append([_ for _ in map(itemgetter(1), g)])

    for i in range(len(range_tuples)):
        range_tuples[i] = (range_tuples[i][0], range_tuples[i][-1]+1)

    return range_tuples

def range_tuples_to_default_flanks(name, value, range_tuples):
    flanks = [Flank(name=name, value=value, start=_[0], stop=_[1]) for _ in range_tuples]
    return flanks

def get_default_label_flanks(flanks: List[Flank], lookup, video_key, db_interventions):
    intervention = get_intervention(video_key, db_interventions)
    n_frames = intervention.metadata.frames_total

    flank_dict = {_: [] for _ in lookup.keys()}
    for flank in flanks:
        flank_dict[flank.name].extend([i for i in range(flank.start, flank.stop)])
    
    add_defaults = defaultdict(list)
    for key in flank_dict.keys():
        flank_dict[key]=set(flank_dict[key])
        for i in range(n_frames):
            if i not in flank_dict[key]:
                add_defaults[key].append(i)
    
    new_flanks = []
    for key in add_defaults.keys():
        default_value = lookup[key]
        if default_value != None:
            add_defaults[key].sort()
            ranges = get_consecutive_ranges(add_defaults[key])
            new_flanks.extend(range_tuples_to_default_flanks(key, lookup[key], ranges))   

    return new_flanks

def binarize_multiclass_array(pred: np.array, choices: List):
    classes = [i for i in range(len(choices))]
    binarized_predictions = label_binarize(pred, classes = classes)
    return binarized_predictions




import os
from ..mongodb.get_objects import get_intervention
import requests
from ..extern.requests import extract_video_intervals

# Base
def get_frame_name(i, prefix, suffix):
    name = "{:0>7d}".format(i)
    if prefix:
        name = prefix+name
    if suffix:
        name = name+suffix
    return name

def get_frame_dir(video_key, base_dir_frames):
    frame_dir = base_dir_frames.joinpath(video_key)
    if not frame_dir.exists():
        os.mkdir(frame_dir)
        os.system(f"chgrp -R lux_tomcat {frame_dir.as_posix()}")
    return frame_dir


def extract_range_by_video_path(
    url, auth, video_path,
    target_dir, intervals, suffix, prefix=None,
    ):
    paths = {}
    skip_intervals = []
    for n_interval, interval in enumerate(intervals):
        _paths = {}
        for i in range(interval["start"], interval["end"]):
            _paths[i] = target_dir.joinpath(get_frame_name(i, prefix, suffix))
        all_exist = False
        for i, _path in _paths.items():
            if not _path.exists():
                all_exist = False
                break
        if all_exist:
            skip_intervals.append(n_interval)
        else:
            paths.update(_paths)
    
    intervals = [_ for i, _ in enumerate(intervals) if not i in skip_intervals]
    _r = {
        "videoPath": video_path.as_posix(),
        "intervals": intervals,
        "imageFolderPath": target_dir.as_posix()
    }
    if prefix:
        _r["prefix"]=prefix
    if suffix:
        _r["extension"]=suffix

    r = extract_video_intervals(_r, url, auth)
    return paths
    

# Pipelines
def extract_frames_for_video_segmentation(video_key, url, auth, db_interventions, base_dir_frames, suffix = ".jpg", prefix = None):
    intervention = get_intervention(video_key, db_interventions)
    video_path = intervention.metadata.path
    extract_dict = {
            "url": url,
            "auth": auth,
            "target_dir": get_frame_dir(video_key, base_dir_frames),
            "intervals": [],
            "prefix": prefix,
            "suffix": suffix,
            "video_path": video_path

        }

    for name, video_segmentation in intervention.video_segments_annotation.items():
        _intervals = [{"start": flank.start, "end": flank.stop} for flank in video_segmentation.value]
        extract_dict["intervals"].extend(_intervals)

    paths = extract_range_by_video_path(**extract_dict)
    return paths


import re
import pandas as pd
import requests
from pathlib import Path
from .fieldnames import *

class Terminology:
    """[summary]
    """
    def __init__(self, cfg: dict, terminology_path: Path = None, terminology_type: int = 0):
        if not terminology_path:
            self.terminology_path = cfg["path_terminology"]
        else:
            self.terminology_path = terminology_path
        self.url = cfg["url_endobox_extreme"]+":"+ cfg["port_webserver"] + "/data"
        self.auth = (cfg["user_webserver"], cfg["password_webserver"])
        self.cfg = cfg
        self.concept_attributes = cfg["terminology_concept_attributes"]
        self.terminology_type = terminology_type
        
        self.terminology_df = pd.read_excel(self.terminology_path)


    def get_concepts(self):
        return [_ for _ in self.concept_attributes.keys()]

    def get_concept_query(self, concept): 
        assert concept in self.concept_attributes

        attributes = self.concept_attributes[concept]
        child_attributes = []
        for attribute in attributes:
            child_attributes.extend(self.get_all_child_attribute_ids(attribute))
        
        attributes.extend(child_attributes)

        query = {
            "$elemMatch": {"$in": attributes}
        }

        return query

    def process_raw_text(self, text: str) -> str:
        """Removes \\r \\n and strips flanking spaces from text.\
            Returns text in lower case.

        Args:
            text (str): text to process

        Returns:
            str: processed text
        """
        text = re.sub(r"\r\n", " ", text)
        text = re.sub(r"\n", " ", text)
        text = re.sub(r"  ", " ", text)
        text = text.strip() 

        return text.lower()

    def post_ontology(self):
        with open(self.terminology_path, "rb") as f:

            r = requests.post(
                self.url + URL_POST_ONTOLOGY,
                auth = self.auth,
                files = {"file": f},
                data = {"type": self.terminology_type})

            return r.status_code
            # assert r.status_code == 201

    def terminology_attribute_id_to_name(self, att_id):
        return self.terminology_df[self.terminology_df["ID"] == att_id]["Attribut"].to_list()[0]

    def get_all_child_attribute_ids(self, att_id, terminology_df = None):
        if not terminology_df:
            terminology_df = self.terminology_df
        return terminology_df[terminology_df["ID"].str.startswith(att_id)].ID.to_list()


    def get_terminology_result(self, text:str, terminology_type: int = None, return_tokens: bool = True, preprocess_text: bool=True):
        if not terminology_type:
            terminology_type = self.terminology_type
        if preprocess_text:
            text = self.process_raw_text(text)

        if return_tokens:
            r = requests.post(
                self.url + URL_TEXT_TO_TOKEN,
                json = {"text": text, "type": terminology_type},
                auth = self.auth
            )
            assert r.status_code == 200
            tokens = r.json()
            tokens[FIELDNAME_TOKENS_VALUE] = [_[FIELDNAME_TOKENS_VALUE] for _ in tokens["tokens"] if not _["modifier"]]
            return tokens

        else:
            r = requests.post(
                self.url + URL_TEXT_TO_XML,
                json = {"text": text, "type": terminology_type},
                auth = self.auth
            )

            assert r.status_code == 200
            return r


    


from ..mongodb.get_objects import get_intervention
from ..datasets.binary_image_classification_ds import BinaryImageClassificationDS
from torch.utils.data import DataLoader
import pandas as pd

def get_intervention_dataloader(video_key, ai_config, db_interventions, db_images, batch_size = 200, num_workers = 12):
    intervention = get_intervention(video_key, db_interventions)
    frame_df = intervention.frame_df()
    cursor = db_images.find(
        {"_id": {"$in": frame_df.id.to_list()}},
        {"metadata": 1}
    )
    images = [{"id": str(_["_id"]), "path": _["metadata"]["path"]} for _ in cursor]
    frame_df = pd.DataFrame.from_records(images)

    dataset = BinaryImageClassificationDS(
        frame_df.path,
        frame_df.id,
        scaling = ai_config.ai_settings.image_scaling,
        training = False
    )
    dataloader = DataLoader(
        dataset = dataset,
        batch_size = batch_size,
        num_workers = num_workers,
        shuffle = False   
    )

    return dataloader


from pydantic import BaseModel, validator, constr
from typing import Dict, Any, List, Tuple, Optional, Union
from pathlib import Path
from pymongo.collection import Collection
from pymongo import MongoClient
from .train_data import TrainDataSamplerSettings
from a_ukw_ml_tools.classes.annotation import (
    BinaryAnnotation,
    MultilabelAnnotation,
    MultichoiceAnnotation
)
from a_ukw_ml_tools.classes.prediction import (
    BinaryPrediction,
    MultilabelPrediction,
    MultichoicePrediction
)

LOOKUP_LABEL_CLASS = {
    "annotation": {
        "binary": BinaryAnnotation,
        "multilabel": MultilabelAnnotation,
        "multiclass": MultichoiceAnnotation
    },
    "prediction": {
        "binary": BinaryPrediction,
        "multilabel": MultilabelPrediction,
        "multiclass": MultichoicePrediction
    }
}

# LOOKUP_MODEL_CLASS = 

class AiSettings(BaseModel):
    version: float
    image_scaling: int
    base_model_name: str

class AiLabelConfig(BaseModel):
    name: str
    label_type: str
    is_compound: bool
    annotation_class: Optional[Union[
        BinaryAnnotation,
        MultilabelAnnotation,
        MultichoiceAnnotation
    ]]
    prediction_class: Optional[Union[
        BinaryPrediction,
        MultilabelPrediction,
        MultichoicePrediction
    ]]
    label_type: str
    choices: List[Union[bool, str]]
    ai_settings: AiSettings
    sampler_settings: TrainDataSamplerSettings

    @validator("annotation_class", always = True)
    def set_annotation_class(cls, v, values):
        v = LOOKUP_LABEL_CLASS["annotation"][values["label_type"]]
        return v

    @validator("prediction_class", always = True)
    def set_prediction_class(cls, v, values):
        v = LOOKUP_LABEL_CLASS["annotation"][values["label_type"]]
        return v

    # def get_primary_choices(self):
    #     """Returns choices without the labels "outside" and "blurry" """

class LabelSettings(BaseModel):
    default_values: Dict[int, Any]
    skip_labels_flank_to_image: List[str]

class BasePaths(BaseModel):
    frames: Path
    models: Path
    train_data: Path

    class Config:
        allow_population_by_field_name = True
        json_encoders = {Path: str}
        schema_extra = {
            "example": {}
        }

class TimeFormat(BaseModel):
    video_extern: str

class Auth(BaseModel):
    labelstudio_token: str
    extern_user: str
    extern_pw: str
    extern_auth: Optional[Tuple[str]]

    @validator("extern_auth", always=True)
    def add_auth(cls, v, values):
        v = (values["extern_user"], values["extern_pw"])
        return v

class WebSettings(BaseModel):
    ip_gpu_server: str
    ip_endobox_extreme: str
    port_extern: str
    port_labelstudio: str
    port_fileserver: str
    url_mongodb: str
    url_extern: Optional[str]

    @validator("url_extern", always=True)
    def add_url_extern(cls, v, values):
        v = "https://" 
        v += values["ip_endobox_extreme"]
        v += ":"
        v += values["port_extern"]
        v += "/data"
        return v

class Databases(BaseModel):
    url_mongodb: str

    mongo_client: Optional[Any]

    @validator("mongo_client", pre = True, always = True)
    def get_client(cls, v, values):
        v = MongoClient(
            values["url_mongodb"],
            connectTimeoutMS=200,
            retryWrites = True
        )
        return v

class Configuration(BaseModel):
    label_settings: LabelSettings
    base_paths: BasePaths
    auth: Auth
    web: WebSettings

    db: Optional[Databases]

    @validator("db", pre = True, always = True)
    def setup_db(cls, v, values):
        v = Databases(url_mongodb = values["web"].url_mongodb)
        return v


    def get_databases(self):
        """
        Returns (db_images, db_interventions, db_test_data, db_train_data)
        """
        mongo_client = self.db.mongo_client
        db_interventions = self.db.mongo_client
        db_interventions = mongo_client.EndoData2.Interventions
        db_images = mongo_client.EndoData2.Images
        db_test_data = mongo_client.EndoData2.TestData
        db_train_data = mongo_client.EndoData2.TrainData

        return db_images, db_interventions, db_test_data, db_train_data

    def get_test_databases(self):
        """
        Returns collections from TestEndoData(db_images, db_interventions, db_test_data, db_train_data)
        """
        mongo_client = self.db.mongo_client
        db_interventions = self.db.mongo_client
        db_interventions = mongo_client.TestEndoData.Interventions
        db_images = mongo_client.TestEndoData.Images
        db_test_data = mongo_client.TestEndoData.TestData
        db_train_data = mongo_client.TestEndoData.TrainData

        return db_images, db_interventions, db_test_data, db_train_data

    def get_extern_tuple(self):
        """Returns (url, auth)"""
        url = self.web.url_extern
        auth = self.auth.extern_auth
        return url, auth


from datetime import datetime as dt

def path_to_timestamp(path, time_format, origin=None):
    name = path.name
    if origin:
        name = name.replace(origin, "")
    try:
        time_str = name.split("_") [-3:-1]
        time_str = "_".join(time_str)
        timestamp = dt.strptime(time_str, time_format)
    except:
        time_str = name.split("_") [-4:-2]
        time_str = "_".join(time_str)
        timestamp = dt.strptime(time_str, time_format)    

    return timestamp

def video_extern_exists_intern(video_id, db_interventions) -> bool:
    _ = db_interventions.find_one({"id_extern": video_id})
    if _: return True
    else: return False

def video_extern_get_intern_id(video_id, db_interventions):
    _ = db_interventions.find_one({"id_extern": video_id})
    if _: return _["_id"]




# Interventions
FIELDNAME_FRAMES = "frames"
FIELDNAME_VIDEO_PATH = "video_path"
FIELDNAME_TOKENS = "tokens"
FIELDNAME_TOKENS_PATHO = "patho"
FIELDNAME_TOKENS_REPORT = "report"
FIELDNAME_TOKENS_VALUE = "value"
FIELDNAME_EXTERNAL_ID = "_id_michael"
FIELDNAME_INTERVENTION_IMAGE_ANNOTATIONS = "image_annotations"
FIELDNAME_INTERVENTION_TYPE = "intervention_type"
FIELDNAME_FREEZES = "freezes"
FIELDNAME_INTERVENTION_DATE = "intervention_date"
FIELDNAME_VIDEO_METADATA = "metadata"
FIELDNAME_FPS = "fps"
FIELDNAME_FRAMES_TOTAL = "frames_total"
FIELDNAME_PATHO_RAW = "patho_raw"
FIELDNAME_REPORT_RAW = "report_raw"
FIELDNAME_INTERVENTION_IMAGE_PREDICTIONS = "image_predictions"
FIELDNAME_VIDEO_DURATION = "duration"
FIELDNAME_N_EXTRACTED_FRAMES = "n_extracted_frames"
FIELDNAME_PREDICTIONS_SMOOTH = "predictions_smooth"
FIELDNAME_PREDICTION_FLANKS = "prediction_flanks"
FIELDNAME_ANNOTATION_FLANKS = "annotation_flanks"
FIELDNAME_FLANK_RANGES = "ranges"
FIELDNAME_FLANK_FRAMES = "frames"
FIELDNAME_VIDEO_ANNOTATION_TIMESTAMP = "latest_video_annotation_date"

# Images
FIELDNAME_INTERVENTION_ID = "intervention_id"
FIELDNAME_LABELS = "labels_new"
FIELDNAME_IMAGE_PATH = "path"
FIELDNAME_FRAME_NUMBER = "n"
FIELDNAME_LABELS_VALIDATED = "labels_validated"
FIELDNAME_LABELS_UNCLEAR = "labels_unclear"
FIELDNAME_IMAGE_TYPE = "image_type"
LABEL_UNCLEAR = "unclear"

# Shared
FIELDNAME_ORIGIN = "origin"
FIELDNAME_VIDEO_KEY = "video_key"

# AI
FIELDNAME_AI_VERSION = "version"
FIELDNAME_IMAGE_SCALING = "image_scaling"
COLNAME_RESULT_TYPE = "type"
RESULT_TYPE_PREDICTION = "prediction"
RESULT_TYPE_ANNOTATION = "annotation"
RESULT_TYPE_PREDICTION_SMOOTH = "prediction_smooth"
COLNAME_PREDICTION_VALUE = "value"
FIELDNAME_PREDICTIONS = "predictions"
COLNAME_RESULT_LABEL = "label"
COLNAME_RESULT_NAME = "name"

# OTHER STRING CONSTANTS
IMAGETYPE_FRAME = "frame"
IMAGETYPE_FREEZE = "freeze"

# TERMINOLOGY
URL_POST_ONTOLOGY = "/PostOntology"
URL_TEXT_TO_TOKEN = "/PostTerminologyTokens"
URL_TEXT_TO_XML = "/PostTerminologyXml"

# PREDICTION_RESULT_DF
COLNAME_AI_NAME = "ai_name"


# CFG

# STATS
"""
Components in the name are not allowed to contain . or ,
Stats Result names for dict are defined as:
{db}.{value_type}.{att1}.{value_1}.{att2}.{value2}.....

if an attribute has more than one value, they are to be comma separated

"""

PREFIX_COUNT = "count"
PREFIX_INTERVENTION = "interventions"
PREFIX_IMAGE = "images"
PREFIX_EXISTS = "exists"

DF_COL_ENTITY = "entity"
DF_COL_VALUE = "value"
DF_COL_VALUE_TYPE = "value_type"
DF_COL_DATE = "date"




def instantiate(config):
    config = config.copy()
    _class = config["_target_"]
    del config["_target_"]
    instance = _class(**config)
    
    return instance

def load_model_trainer(trainer_settings, db_train_data, db_test_data):
    pass 


# Test Data
TD_NAME = "name"
TD_IDS = "ids"

# Images
IMG_INTERVENTION_ID = "intervention_id"
IMG_INTERVENTION_ID = "intervention_id"
IMG_ANNOTATIONS = "annotations"
IMG_FRAME_NUMBER = "frame_number"
IMG_IS_FRAME = "is_frame"

# Shared
ORIGIN = "origin"
METADATA = "metadata"



from typing import Dict, List
from ..classes.annotation import VideoSegmentationAnnotation
from bson import ObjectId
from .get_objects import get_intervention
from ..classes.image import Image
from ..classes.intervention import Intervention
from pymongo.collection import Collection

def update_video_segments_annotation(
    video_key,
    video_segmentation: Dict[str, VideoSegmentationAnnotation],
    db_interventions
    ):
    video_segmentation = {key: value.dict() for key, value in video_segmentation.items()}
    r = db_interventions.update_one({"video_key": video_key},
    {"$set": {"video_segments_annotation": video_segmentation}})

    return r

def insert_frame(image:Image, db_images:Collection):
    assert not image.id
    r = db_images.insert_one(image.to_dict())
    return r

def insert_new_intervention(intervention:Intervention, db_interventions: Collection):
    assert not intervention.id
    r = db_interventions.insert_one(intervention.to_dict())
    return r

def add_frames_to_framedict(video_key, inserted_ids: Dict[int, ObjectId], db_interventions: Collection):
    intervention = get_intervention(video_key, db_interventions)
    
    for n_frame, _id in inserted_ids.items():
        assert n_frame not in intervention.frames
        db_interventions.update_one({"video_key": video_key}, {"$set": {f"frames.{n_frame}": _id}})

def update_train_dataset(name, train_set_db, db_train_data):
    db_train_data.update_one({"name": name}, {"$set": train_set_db.dict()}, upsert = True)

def update_many_frame_is_extracted(value: bool, id_list:List[ObjectId], db_images):
    db_images.update_many({"_id": {"$in": id_list}}, {"$set": {"metadata.is_extracted": value}})


from ..extern.requests import get_extern_annotations
from ..mongodb.data_import import (
    filter_extern_annotation_by_latest_date,
    get_video_segmentation_extern,
    generate_db_images_for_frame_extraction,
    get_image_updates_for_video_segmentation
)
from ..media.frame_extraction import extract_frames_for_video_segmentation
from pymongo.collection import Collection
from ..mongodb.get_objects import get_intervention
from ..mongodb.update import (
    update_video_segments_annotation,
    insert_frame, 
    add_frames_to_framedict,
)
from ..mongodb.base import filter_existing_frames_by_intervention
from ..classes.config import LabelSettings

def get_new_annotation_videos(url:str, auth: str, db_interventions:Collection):
    annotated = get_extern_annotations(url, auth)
    print("Annotated Videos:", len(annotated))
    annotated_new = filter_extern_annotation_by_latest_date(annotated, db_interventions)
    print(len(annotated_new), "Videos have new annotations")
    return annotated_new

def import_new_video_annotation(annotated_video, label_group_settings:LabelSettings, url, auth, db_interventions):
    extern_latest_annotation = annotated_video.date
    video_key = annotated_video.video_key

    video_segmentation = get_video_segmentation_extern(
        video_key, extern_latest_annotation,
        label_group_settings.default_values,
        url,
        auth,
        db_interventions
    )
    
    r = update_video_segments_annotation(video_key, video_segmentation, db_interventions)
    return r
    

def extract_and_generate_frames(annotated_video, base_dir_frames, url, auth, db_interventions, db_images):
    video_key = annotated_video.video_key
    paths = extract_frames_for_video_segmentation(
        video_key,
        url,
        auth,
        db_interventions,
        base_dir_frames
    )
    paths = filter_existing_frames_by_intervention(video_key, paths, db_interventions)
    images = generate_db_images_for_frame_extraction(video_key, paths, db_interventions)
    inserted_ids = {}
    for image in images:
        r = insert_frame(image, db_images)
        inserted_ids[image.metadata.frame_number] = r.inserted_id
    add_frames_to_framedict(video_key, inserted_ids, db_interventions)
    return inserted_ids

def set_image_label_updates_from_video_segmentations(
        annotated_video,
        ai_config_dict,
        db_interventions,
        db_images,
        set_in_db = True
    ):
    """
    Executes Updates and sets
    Returns Tuples like ({"_id": ObjectId("asd"), {"$set": {"annotations.name": Annotation}}})
    """
    video_key = annotated_video.video_key
    timestamp = annotated_video.date
    updates = get_image_updates_for_video_segmentation(
        video_key, db_interventions, ai_config_dict
    )
    if set_in_db:
        for update in updates:
            db_images.update_one(update[0], update[1])
    
    return updates


from bson.objectid import ObjectId
from pydantic import BaseModel, validator
from typing import Union, List, Optional, Set
from pathlib import Path
import json
import numpy as np
import pandas as pd
from .base import PyObjectId
from ..labels.strat_group_split import GradientSolver, generate_counts, create_initial_solution
from sklearn.utils import shuffle


class TrainDataSamplerSettings(BaseModel):
    min_frame_diff: int
    test_sets: List[str]
    exclude_origins: List[str]    

class TrainDataBase(BaseModel):
    name: str
    prediction_type: str    
    choices: List[str]
    weights: Optional[List[float]]

    df_columns: Set = {
            "labels",
            "paths",
            "origins",
            "intervention_ids",
            "image_ids",
            "is_val"
        }

    class Config:
        allow_population_by_field_name = True
        # arbitrary_types_allowed = True
        json_encoders = {Path: str, ObjectId: str, Set: list}
        schema_extra = {
            "example": {}
        }

    def get_df_path(self, base_path_train_data):
        return base_path_train_data.joinpath(self.name).with_suffix(".csv")

class TrainDataDb(TrainDataBase):
    path: Path

    class Config:
        allow_population_by_field_name = True
        # arbitrary_types_allowed = True
        json_encoders = {Path: str, ObjectId: str, Set: list}
        schema_extra = {
            "example": {}
        }

    @validator("path")
    def path_to_str(cls, v):
        return v.as_posix()

    @validator("df_columns")
    def df_columns_to_list(cls, v):
        return list(v)

    def to_dict_intern(self):
        _dict = self.dict()
        df = pd.read_csv(self.path)
        df_dict = df.to_dict(orient= "list")
        _dict.update(df_dict)
        return _dict


class TrainData(TrainDataBase):
    labels: Union[List[bool], List[int], List[List[int]]]
    origins: List[str]
    intervention_ids: List[PyObjectId]
    paths: List[Path]
    image_ids: List[PyObjectId]
    is_val: Optional[List[bool]]

    def parsed_dict(self):
        return json.loads(self.json(by_alias = True, exclude_none = True))

    def n_classes(self):
        return len(self.choices)

    def get_weights(self, target = None):
        if not target :
            target = self.labels
        class_sample_count = np.array(
        [len(np.where(target == t)[0]) for t in np.unique(target)])
        weight = 1. / class_sample_count
        samples_weight = np.array([weight[t] for t in target])
        return samples_weight

    def get_df(self):
        _dict = self.dict(
            include = self.df_columns
        )
        df = pd.DataFrame.from_dict(_dict)
        return df

    def save_df(self, base_path_train_data):
        path = self.get_df_path(base_path_train_data)
        df = self.get_df()
        df.to_csv(path, index = False)
        return path

    def to_db_dict(self, base_path_train_data):
        db_dict = self.dict(exclude = self.df_columns)
        db_dict["path"] = self.get_df_path(base_path_train_data)
        return db_dict

    def group_strat_split(
        self,
        test_size = 0.1,
        return_val_ids = False,
        max_empty_iterations = 500,
        max_intensity_iterations = 100,
        min_cost = 100,
        verbose = False   
        ):

        train_df = self.get_df()
        counts = train_df.groupby("intervention_ids")["labels"].value_counts().unstack().fillna(0)
        sample_cnt = counts.to_numpy()

        solution_arr = create_initial_solution(sample_cnt, test_size)
        g_solver = GradientSolver(sample_cnt, solution_arr, test_size)
        solution = g_solver.solve(min_cost, max_empty_iterations, max_intensity_iterations, verbose=False)
        solution_array = solution.index

        val_ids = counts.index[solution_array]
        train_ids = counts.index[~solution_array]

        if return_val_ids:
            return val_ids

        _train_df = train_df[train_df["intervention_ids"].isin(train_ids)]
        _train_df = shuffle(_train_df)
        _train_df.reset_index(inplace=True, drop=True)

        _val_df = train_df[train_df["intervention_ids"].isin(val_ids)]
        _val_df = shuffle(_val_df)
        _val_df.reset_index(inplace=True, drop=True)

        x_train = np.array(_train_df["paths"].astype(str))
        x_val = np.array(_val_df["paths"].astype(str))

        y_train = np.array(_train_df["labels"])
        y_val = np.array(_val_df["labels"])

        if verbose:
            print("Length Validation:", len(_val_df))
            print("Length Train:", len(_train_df))
            print(f"Percentage Val Data (Target: {test_size}):", len(_val_df) / (len(_val_df)+len(_train_df)))
            train_counts = _train_df["labels"].value_counts().sort_index()
            val_counts = _val_df["labels"].value_counts().sort_index()
            
            print("Class counts Validation")
            print(val_counts)
            print("Class counts Training")
            print(train_counts)
            print("Validation Class Percentages of Total")
            print(val_counts / (train_counts + val_counts))

        return x_train, x_val, y_train, y_val



import requests
from ..classes.extern import ExternAnnotatedVideo, VideoExtern
from typing import List, Tuple
from .base import get_extern_ids, get_base_image_dict_from_intervention
from ..classes.intervention import Intervention
from ..classes.annotation import VideoSegmentationAnnotation
from ..classes.image import Image
from ..extern.requests import (
    get_extern_annotations,
    get_extern_video_annotation,
    get_extern_interventions
)
from ..extern.conversions import (
    annotations_by_label_group,
    get_flank_dict,
    get_labels_with_test_annotation
)
from collections import defaultdict
from .get_objects import get_intervention
from ..labels.conversions import get_default_label_flanks

def filter_duplicate_keys(interventions):
    _duplicates = []
    video_keys = []
    names = [_.path.name for _ in interventions]
    for _ in names:
        if _ in video_keys:
            _duplicates.append(_)
        else:
            video_keys.append(_)

    duplicates = [(_.path.name, _.id_extern) for _ in interventions if _.path.name in _duplicates]
    interventions = [_ for _ in interventions if not _.path.name in _duplicates]
    print(f"Dropped videos for {len(duplicates)} unique video_keys")

    return interventions, duplicates

def filter_extern_annotation_by_latest_date(annotated_video_list: List[ExternAnnotatedVideo], db_interventions):
    new_annotated_videos = []
    for annotated_video in annotated_video_list:
        intervention = get_intervention(annotated_video.video_key, db_interventions)
        if intervention.video_segments_annotation:
            dates = [annotation.date for name, annotation in intervention.video_segments_annotation.items()]
        _is_new = False
        for date in dates:
            if annotated_video.date > date:
                _is_new = True
                break
        if _is_new:
            new_annotated_videos.append(annotated_video)

    return new_annotated_videos

def get_new_extern_interventions(
        url_extern: str, auth_extern: Tuple[str], db_interventions
    ) -> Tuple[List[VideoExtern]]:
    """Calls get_extern_interventions, filters list of VideoExtern\
    Objects for duplicate keys and for extern_ids which already exist in\
    the given db.

    Args:
        url_extern (str): e.g. "https://10.235.14.33:8443/data"
        auth_extern (str): e.g. ("user", "password")
        db_interventions ([type]): 

    Returns:
        Tuple[List[VideoExtern]]: Returns Tuple containing 3 Lists:\
            interventions, failed conversions, duplicates
    """
    videos = get_extern_interventions(url_extern, auth_extern)

    videos, duplicates = filter_duplicate_keys(videos)

    existing_extern_video_ids = get_extern_ids(db_interventions)
    videos = [_ for _ in videos if _.id_extern not in existing_extern_video_ids]
    interventions = []
    failed = []
    for _ in videos:
        try:
            interventions.append(Intervention(**_.to_intervention_dict(db_interventions)))
        except:
            failed.append(_)
    print(f"Failed to convert {len(failed)} interventions")
    
    return interventions, failed, duplicates

def get_video_segmentation_extern(video_key, latest_session, default_values_dict, url, auth, db_interventions):
    video_annotations = get_extern_video_annotation(video_key, url, auth)

    video_annotation_dict = annotations_by_label_group(video_annotations)
    flanks, lookup = get_flank_dict(video_annotation_dict, default_values_dict)
    flanks["test"] = get_default_label_flanks(flanks["test"], lookup, video_key, db_interventions)
    flanks = flanks["train"] + flanks["test"]
    video_segmentation = {}
    for flank in flanks:
        name = flank.name
        if not name in video_segmentation:
            video_segmentation[name] = VideoSegmentationAnnotation(
                source = "web_annotation_flanks",
                annotator_id=0,
                date = latest_session,
                name = name,
                value = []
                )
        video_segmentation[name].value.append(flank)

    return video_segmentation

def generate_db_images_for_frame_extraction(video_key, paths, db_interventions):
    intervention = get_intervention(video_key, db_interventions)
    base_image_dict = get_base_image_dict_from_intervention(intervention)
    images = []

    for n_frame, path in paths.items():
        image_dict = base_image_dict.copy()
        image_dict["metadata"] = {
            "frame_number": n_frame,
            "path": path,
            "is_extracted": False,
            "is_frame": True
        }
        images.append(Image(**image_dict))

    return images

from ..classes.config import AiLabelConfig

def video_segmentation_to_image_labels(
    video_segmentation_annotation:VideoSegmentationAnnotation,
    ai_label_config_dict,
    skip_labels:str,
    is_annotation:bool = True
    ):
    
    annotator_id = video_segmentation_annotation.annotator_id
    flanks = video_segmentation_annotation.value
    date = video_segmentation_annotation.date

    if is_annotation:
        _class_selector = "annotation_class"
    else:
        _class_selector = "prediction_class"

    image_labels = defaultdict(list)
    for flank in flanks:
        if flank.name in skip_labels:
            continue
        ai_label_config = ai_label_config_dict[flank.name]
        ai_label_config = AiLabelConfig(**ai_label_config)
        _class = ai_label_config.dict()[_class_selector]
        assert flank.value in ai_label_config.choices

        for i in range(flank.start, flank.stop):
            image_labels[i].append(_class(
                source = "web_annotation_flanks",
                annotator_id=annotator_id,
                date = date,
                choices = ai_label_config.choices,
                name = flank.name,
                value = flank.value
            ))
    return image_labels


def image_label_dict_to_updates(video_key, image_label_dict, db_interventions):
    intervention = get_intervention(video_key, db_interventions)
    updates = []
    for n, annotations in image_label_dict.items():
        image_id = intervention.frames[n]
        _set = {}
        for annotation in annotations:
            _set[f"annotations.{annotation.name}"] = annotation.dict()
        
        
        _update = ({"_id": image_id}, {"$set": _set})
        updates.append(_update)

    return updates   


def get_image_updates_for_video_segmentation(
        video_key, 
        db_interventions,
        ai_label_config_dict,
        skip_labels = ["withdrawal"]
    ):
    intervention = get_intervention(video_key, db_interventions)
    video_segmentation_annotation = intervention.video_segments_annotation
    image_labels = defaultdict(list)
    for name, segmentation in video_segmentation_annotation.items():
        _image_labels = video_segmentation_to_image_labels(
            segmentation,
            ai_label_config_dict,
            is_annotation=True,
            skip_labels=skip_labels
        )
        for n_frame, value in _image_labels.items():
            image_labels[n_frame].extend(value)

    updates = image_label_dict_to_updates(video_key, image_labels, db_interventions)

    return updates


from ..mongodb.data_import import get_new_extern_interventions
from ..classes.intervention import Intervention
from ..mongodb.update import insert_new_intervention
from typing import List, Tuple
from pymongo.collection import Collection

def import_new_extern_interventions(url:str, auth: Tuple[str], db_interventions) -> List:
    interventions, failed, duplicates = get_new_extern_interventions(url, auth, db_interventions)
    video_keys = [_ for _ in db_interventions.distinct("video_key")]
    inserted_ids = []
    for intervention in interventions:
        intervention = Intervention(**intervention.to_intervention_dict(db_interventions))
        assert not intervention.id
        assert intervention.video_key not in video_keys

        r = insert_new_intervention(intervention, db_interventions)
        inserted_ids.append(r.inserted_id)

    return inserted_ids




from pathlib import Path
import json
import warnings
from bson.objectid import ObjectId
from pymongo import MongoClient, aggregation, UpdateOne
from .db_interventions import DbInterventions
from .db_images import DbImages
from .db_tests import DbTests
from .intervention import Intervention
from .utils import *
from .fastcat_annotation import FastCatAnnotation
from .labelstudio import Labelstudio
from typing import List, Tuple
from collections import Counter
from .fieldnames import *
import pandas as pd
from tqdm import tqdm
from .db_extern import DbExtern
from datetime import datetime as dt


class Db:
    '''
    Add Documentation
    '''
    def __init__(self, path_config: Path):
        with open(path_config, "r", encoding = "utf-8") as f:
            self.cfg = json.load(f)
        
        # URLS
        self.url_mongodb = self.cfg["uri"]
        self.url_endobox_extreme = self.cfg["url_endobox_extreme"]
        self.url_gpu_server = self.cfg["url_gpu_server"]
        port_fileserver = self.cfg["port_fileserver"]
        port_labelstudio = self.cfg["port_labelstudio"]

        self.fileserver_prefix = self.url_endobox_extreme + ":" + port_fileserver + "/"
        self.labelstudio_prefix = self.url_endobox_extreme + ":" + port_labelstudio + "/api/"
        self.labelstudio_token = self.cfg["labelstudio_token"]
        self.streamlit_config = self.cfg["streamlit_config"]
        
        # Paths
        base_path_stats = Path(self.cfg["base_path_stats"])
        self.paths = {
            "models": Path(self.cfg["base_path_models"]),
            "stats": {
                "base": base_path_stats,
                "label_count": base_path_stats.joinpath("labelcount.json"),
                "label_by_origin": base_path_stats.joinpath("label_by_origin_count.json"),
                "label_by_video": base_path_stats.joinpath("video_origin_count.json")
            },
            "frames": Path(self.cfg["base_path_frames"]),
            "ls_configs": Path(self.cfg["path_label_configs"])
        }
        
        # Misc
        self.fast_cat_allowed_labels = self.cfg["fast_cat_allowed_labels"]
        self.skip_frame_factor = self.cfg["skip_frame_factor"]
        self.ai_model_settings = self.cfg["models"]
        self.exclude_origins_for_frame_diff_filter = self.cfg["exclude_origins_for_frame_diff_filter"]
        self.db_extern_time_format = self.cfg["db_extern_date_pattern"]
        self.db_extern_user_priority = self.cfg["db_extern_annotation_user_priority"]
        self.db_extern_test_data_annotation_group_ids = self.cfg["db_extern_test_data_annotation_group_ids"]

        # Setup Mongo Db connection
        self.mongo_client = MongoClient(self.url_mongodb, connectTimeoutMS=200, retryWrites=True)
        self.db_images = DbImages(self.mongo_client.EndoData.images, self.cfg)
        self.db_interventions = DbInterventions(self.mongo_client.EndoData.interventions, self.cfg)
        self.db_tests = DbTests(self.mongo_client.EndoData.test_data, self.db_images, self.db_interventions, self.cfg)
        self.db_stats = self.mongo_client.EndoData.stats

        # Setup db_extern
        self.db_extern = DbExtern(self.cfg)

        # Setup Labelstudio connection
        self.labelstudio = Labelstudio(
            ls_url=self.labelstudio_prefix,
            ls_token = self.labelstudio_token,
            ls_config_path=self.paths["ls_configs"],
            cfg = self.cfg
            )

    def import_new_extern_interventions(self):
        new_interventions = self.db_extern.get_extern_interventions()
        new_interventions = [self.db_extern.convert_intervention_external_to_internal(_) for _ in new_interventions]
        print(f"Found {len(new_interventions)} interventions in external db")
        imported = 0
        failed = []
        for intervention in tqdm(new_interventions):
            try:
                _ = self.db_interventions.create_new_from_external(intervention)
            except:
                _ = None
                failed.append(intervention)

            if _:
                imported += 1
        print(f"Imported {imported} new interventions")
        print(f"{len(failed)} failed")

        return failed

    def calculate_stats(self):
        stats = self.db_interventions.calculate_stats() + self.db_images.calculate_stats()
        self.db_stats.insert_one({"type": "count_data", "data": stats})

    def update_all(self):
        log = {}
        print(dt.now(), "Import new from Extern")
        log["import_extern"] = self.import_new_extern_interventions()
        print(dt.now(), "Calculate Stats")
        self.calculate_stats()
        print(dt.now(), "Update Video Metadata")
        log["metadata"] = self.update_all_video_metadata(only_new = True)
        print(dt.now(), "Update Interventions' Image Annotations")
        self.update_all_intervention_image_annotations()
        print(dt.now(), "Update Interventions' Image Predictions")
        self.update_all_intervention_image_predictions()
        print(dt.now(), "Applying Terminology")
        self.db_interventions.apply_terminology(set_tokens_in_db = True, return_tokens = True)
        return log

    def update_all_intervention_image_annotations(self):
        interventions = self.db_interventions.db.find({})
        interventions = [_ for _ in interventions]
        
        for intervention in tqdm(interventions):
            intervention = Intervention(intervention["_id"], self.db_images.db, self.db_interventions.db, self.cfg)
            intervention.update_image_annotations()

    def update_all_intervention_image_predictions(self):
        interventions = self.db_interventions.db.find({})
        interventions = [_ for _ in interventions]
        
        for intervention in tqdm(interventions):
            intervention = Intervention(intervention["_id"], self.db_images.db, self.db_interventions.db, self.cfg)
            intervention.update_image_predictions()

    def update_all_video_metadata(self, only_new: bool = True):
        q = {FIELDNAME_VIDEO_PATH: {"$exists": True, "$nin": ["", None]}}
        if only_new:
            q[FIELDNAME_VIDEO_METADATA] = {"$exists": False}

        interventions = self.db_interventions.db.aggregate([{"$match":q}])
        interventions = [_ for _ in interventions]

        failed = []

        for intervention in tqdm(interventions):
            try:
                intervention = Intervention(intervention["_id"], self.db_images.db, self.db_interventions.db, self.cfg)
                meta = intervention.get_video_metadata()
                self.db_interventions.db.update_one({"_id": intervention._id}, {"$set": {FIELDNAME_VIDEO_METADATA: meta}})
            except:
                failed.append(str(intervention._id))

        if failed:
            print(f"Could not create metadata for {len(failed)} videos")

        return failed

    def get_stats(self, as_df: bool = True):
        _records = self.db_stats.find({"type": "count_data"})
        records = []
        for record in _records:
            records.extend((record["data"]))

        if as_df:
            return pd.DataFrame.from_records(records)
        else:
            return records 

    def get_metadata_for_train_data(self, metadata, label, value, extend_conditions = None, extend_agg = None):
        # assert label not in metadata
        metadata = metadata.copy()

        test_intervention_ids = self.db_tests.get_test_data_interventions(label) 
        aggregation = get_train_image_query(label, value, test_intervention_ids, extend_conditions, extend_agg)

        aggregation.append(
            {
                "$group": {
                    "_id": {
                        FIELDNAME_ORIGIN: "$"+FIELDNAME_ORIGIN,
                        FIELDNAME_INTERVENTION_ID: "$"+FIELDNAME_INTERVENTION_ID,
                        label: f"${FIELDNAME_LABELS}.{label}"
                    },
                    "count": {"$sum":1}
                }
            }
        )
        _counts = self.db_images.db.aggregate(aggregation)
        origin_count = defaultdict(generate_list_default_dict)

        for _count in _counts:
            origin = _count["_id"][FIELDNAME_ORIGIN]
            intervention = str(_count["_id"][FIELDNAME_INTERVENTION_ID])
            value = _count["_id"][label]
            origin_count[origin][intervention].append(_count["count"])

        for origin in origin_count.keys():
            for intervention in origin_count[origin].keys():
                count = origin_count[origin][intervention]
                origin_count[origin][intervention] = count

        aggregation = get_train_image_query(label, value, test_intervention_ids, extend_conditions, extend_agg)
        aggregation.append(
            {
                "$group": {
                    "_id": {
                        FIELDNAME_ORIGIN: "$"+FIELDNAME_ORIGIN,
                        # FIELDNAME_INTERVENTION_ID: "$"+FIELDNAME_INTERVENTION_ID,
                        label: f"${FIELDNAME_LABELS}.{label}"
                    },
                    "count": {"$sum":1}
                }
            }
        )
        _counts = self.db_images.db.aggregate(aggregation)
        origin_total = {_["_id"][FIELDNAME_ORIGIN]: _["count"] for _ in _counts}


        intervention_dates = {}
        for origin in origin_count.keys():
            _intervention_ids = [ObjectId(_) for _ in origin_count[origin].keys()]
            r = self.db_interventions.db.aggregate([
                {"$match": {"_id": {"$in": _intervention_ids}}},
                {"$group": {
                    "_id": None,
                    "min_date": {"$min": "$"+FIELDNAME_INTERVENTION_DATE},
                    "max_date": {"$max": "$"+FIELDNAME_INTERVENTION_DATE}
                }}
            ]).next()
            if r["min_date"]:
                min_date = dt.strftime(r["min_date"], "%Y-%-m-%d")
                max_date = dt.strftime(r["max_date"], "%Y-%-m-%d")
            else:
                min_date = None
                max_date = None

            intervention_dates[origin] = (min_date,max_date)
        
        n_interventions_origin = {origin: len(list(origin_count[origin].keys())) for origin in origin_count.keys()}
        metadata[label][value] = {
            "origin_total_interventions": n_interventions_origin,
            "origin_total_images": origin_total,
            "intervention_dates": intervention_dates,
            "origin_count": origin_count,
        }

        return metadata

    def get_train_df(self, label: str, min_frame_diff: int = None, verbose: bool = False, return_meta: bool = False):
        settings = self.ai_model_settings[label]
        if return_meta:
            metadata = defaultdict(dict)

        if settings["prediction_type"] == "binary":
            pos_imgs = self.get_train_img_list(label, True, min_frame_diff, verbose)              
            n_pos = len(pos_imgs)
            if return_meta:
                metadata = self.get_metadata_for_train_data(metadata, label, True)
                metadata[label][True]["n_images"] = n_pos

            neg_imgs = self.get_train_img_list(label, False, min_frame_diff, verbose)
            if return_meta:
                metadata = self.get_metadata_for_train_data(metadata, label, False)
                metadata[label][False]["n_images"] = len(neg_imgs)

            for _label, _multiplier in settings["neg_label_list"].items():
                if _label == "normal":
                    ext_conditions = [{f"{FIELDNAME_LABELS}.{label}": {"$exists": False}}]

                    extend_agg = [{"$sample": {"size": int(_multiplier * n_pos)}}]
                    _neg_imgs = self.get_train_img_list(_label, True, min_frame_diff, extend_conditions=ext_conditions, extend_agg = extend_agg)
                    neg_imgs.extend(_neg_imgs)


                elif self.ai_model_settings[_label]["prediction_type"] == "binary":
                    ext_conditions = [{f"{FIELDNAME_LABELS}.{label}": {"$exists": False}}]

                    extend_agg = [{"$sample": {"size": int(_multiplier * n_pos)}}]
                    _neg_imgs = self.get_train_img_list(_label, True, min_frame_diff, extend_conditions=ext_conditions, extend_agg = extend_agg)
                    neg_imgs.extend(_neg_imgs)
                
                else:
                    assert 0 == 1

                if return_meta:
                    metadata = self.get_metadata_for_train_data(metadata, _label, True, ext_conditions, extend_agg)
                    metadata[label][False]["n_images"] = len(neg_imgs)

            df_dict = {
                "file_path": [_["path"] for _ in pos_imgs],
                "label": [1 for _ in pos_imgs]
                }

            df_dict["file_path"].extend([_["path"] for _ in neg_imgs])
            df_dict["label"].extend([0 for _ in neg_imgs])

            label_df = pd.DataFrame.from_dict(df_dict).drop_duplicates()

            if return_meta:
                return label_df, metadata
            else: 
                return label_df

    def import_fast_cat_annotation(self, video_key: str, annotation_path: Path):
        intervention = [_ for _ in self.db_interventions.db.find({FIELDNAME_VIDEO_KEY: video_key})]
        assert len(intervention) == 1
        assert annotation_path.exists()
        intervention = intervention[0]
        annotation = FastCatAnnotation(annotation_path, self.cfg)
        intervention.add_frame_labels(annotation.labels) 

    def get_intervention_frame_labels(self, intervention_id: ObjectId = None):
        if intervention_id:
            interventions = self.db_interventions.db.find({"_id": intervention_id})
        else:
            interventions = self.db_interventions.db.find({"frames": {"$exists": True}})
        
        intervention_label_dict = {}

        for intervention in interventions:
            intervention = Intervention(intervention["_id"], self.db_images.db, self.db_interventions.db, self.cfg)
            unique_labels = intervention.get_unique_frame_labels()
            intervention_label_dict[intervention["_id"]] = unique_labels
            self.db_interventions.db.update_one({"_id": intervention["_id"]}, {"$set": {FIELDNAME_LABELS_FRAMES: unique_labels}})

        return intervention_label_dict

    def get_train_img_list(self, label: str, value, min_frame_diff: int = None, extend_conditions: List = None, extend_agg: List = None, verbose: bool = False):
        test_intervention_ids = self.db_tests.get_test_data_interventions(label) 
        imgs_true = self.db_images.get_train_images(label, value, test_intervention_ids, extend_conditions, extend_agg, as_list = True)

        if min_frame_diff:
            imgs_true = self.filter_images_by_frame_diff(imgs_true, min_frame_diff)

        return imgs_true       

    def filter_images_by_frame_diff(self, img_list: List[dict], min_frame_diff: int = 10):
        filtered_images_dict = defaultdict(list)

        for img in img_list:
            intervention_id = img[FIELDNAME_INTERVENTION_ID]

            if img[FIELDNAME_ORIGIN] in self.exclude_origins_for_frame_diff_filter:
                filtered_images_dict[intervention_id].append(img)
                continue

            append = True
            n_frame = img[FIELDNAME_FRAME_NUMBER]
            
            if filtered_images_dict[intervention_id]:
                for _img in filtered_images_dict[intervention_id]:
                    frame_diff = abs(_img[FIELDNAME_FRAME_NUMBER] - n_frame)
                    if frame_diff < min_frame_diff:
                        append = False
                        break
            
            if append:
                filtered_images_dict[intervention_id].append(img)

        filtered_images = []
        for key, value in filtered_images_dict.items():
            filtered_images.extend(value)
        
        return filtered_images

    def get_videos_with_new_annotations(self):
        annotations = self.db_extern.get_extern_annotations()
        annotations = [
            _ for _ in annotations if 
            check_extern_video_annotation_timestamp(
                _["videoName"], _["lastAnnotationSession"], self.db_interventions.db)
            ]
        
        return annotations

    def update_intervention_annotations_from_extern(self, video_key, str_timestamp, time_format = None, annotation_priority_user_id = None):
        if not time_format:
            time_format = self.db_extern_time_format
        if not annotation_priority_user_id:
            annotation_priority_user_id = self.db_extern_user_priority

        timestamp = dt.strptime(str_timestamp, time_format)

        intervention = Intervention(video_key, self.db_images.db, self.db_interventions.db, self.cfg, True)
        annotations = self.db_extern.get_extern_video_annotation(video_key)

        video_annotation_dict = generate_db_extern_video_annotation_dict(annotations, time_format)
        video_annotations = select_relevant_extern_video_annotations(video_annotation_dict, annotation_priority_user_id)
        test_label_groups = []
        for _ in video_annotations:
            if str(_["label_group_id"]) in self.db_extern_test_data_annotation_group_ids:
                test_label_groups.append(
                    self.db_extern_test_data_annotation_group_ids[str(_["label_group_id"])]
                )

        if test_label_groups:
            for test_label_group in test_label_groups:
                video_annotations = add_false_labels_for_test_data_annotations(
                    intervention.intervention, video_annotations, test_label_group
                )

        video_flanks = defaultdict(dict)
        for _annotation in video_annotations:
            _label = _annotation["label"]
            _value = str(_annotation["value"])
            if _value not in video_flanks[_label]:
                video_flanks[_label][_value] = []
            video_flanks[_label][_value].append(_annotation["flank"])
        # video_annotations

        # For Michael Frame Extraction, just extract flanks
        frames = []
        for _annotation in video_annotations:
            frames.extend([_ for _ in range(_annotation["flank"][0], _annotation["flank"][1])])
        intervention.extract_frames(frames)

        updates = []
        for flank in video_annotations:
            flank_updates = [
                UpdateOne(
                    {"_id": intervention.intervention[FIELDNAME_FRAMES][str(_)]},
                    {
                        "$set": {
                            f"{FIELDNAME_LABELS}.{flank['label']}": flank["value"]
                        }
                    }
                )
                for _ in range(flank["flank"][0], flank["flank"][1]-2)
            ]

            updates.extend(flank_updates)
        r = self.db_images.db.bulk_write(updates)
        r = self.db_interventions.db.update_one(
            {"_id": intervention._id},
            {"$set": {FIELDNAME_ANNOTATION_FLANKS:video_flanks}})

        intervention.update_image_annotations()

        # Set Latest intervention_date
        self.db_interventions.db.update_one({"_id": intervention._id}, {"$set": {FIELDNAME_VIDEO_ANNOTATION_TIMESTAMP: timestamp}})

        return video_annotations



    # Validation
    def validate_intervention_id_of_frames(self) -> tuple((dict, List[ObjectId])):
        interventions = [_ for _ in self.db_interventions.db.find(field_exists_query(FIELDNAME_FRAMES))]
        wrong_ids = defaultdict(list)
        not_all_frames_exist = []

        for intervention in tqdm(interventions):
            frame_ids = [_id for n_frame, _id in intervention[FIELDNAME_FRAMES].items()]
            images = self.db_images.db.find(
                fieldvalue_in_list_query("_id", frame_ids)
            )

            n_frames_extracted = len([_ for _ in intervention[FIELDNAME_FRAMES].keys()])
            n_frames_found = 0
            
            for image in images:
                n_frames_found += 1
                if not image[FIELDNAME_INTERVENTION_ID] == intervention["_id"]:
                    wrong_ids[intervention["_id"]].append(image["_id"])

            if n_frames_found != n_frames_extracted:
                not_all_frames_exist.append(intervention["_id"])

        if wrong_ids:
            warnings.warn("Interventions have frames whose intervention id doesnt match")

        if not_all_frames_exist:
            warnings.warn("Interventions have extracted frames which were not found")

        return (wrong_ids, not_all_frames_exist)            

    def validate_all_intervention_ids_exist(self) -> List:
        _ids = self.db_images.db.distinct(FIELDNAME_INTERVENTION_ID)
        interventions = [_ for _ in self.db_interventions.db.find(fieldvalue_in_list_query("_id", _ids))]

        if not len(_ids) == len(interventions):
            warnings.warn("Image DB points to interventions which were not found")

    def set_image_type(self):
        interventions = self.db_interventions.get_interventions_with_frames()
        for intervention in tqdm(interventions):
            img_ids = [_ for _ in intervention[FIELDNAME_FRAMES].keys()]
            self.db_images.db.update_many(fieldvalue_in_list_query("_id", img_ids), {"$set": {FIELDNAME_IMAGE_TYPE: IMAGETYPE_FRAME}})

        interventions = self.db_interventions.get_interventions_with_freezes()
        for intervention in tqdm(interventions):
            img_ids = [_ for _ in intervention[FIELDNAME_FREEZES].keys()]
            self.db_images.db.update_many(fieldvalue_in_list_query("_id", img_ids), {"$set": {FIELDNAME_IMAGE_TYPE: IMAGETYPE_FREEZE}})

    def validate_all(self):
        validation_errors = {}
        print(dt.now(), "Validate Video Keys")
        validation_errors["video_keys"] = self.db_interventions.validate_video_keys()
        print(dt.now(), "Validate video_paths")
        validation_errors["video_paths"] = self.db_interventions.validate_video_paths()
        print(dt.now(), "Validate frame_dirs")
        validation_errors["frame_dirs"] = self.db_interventions.validate_frame_dirs()
        print(dt.now(), "Validate image_paths")
        validation_errors["image_paths"] = self.db_images.validate_image_paths()
        print(dt.now(), "Validate intervention_ids")
        validation_errors["intervention_ids"] = self.db_images.validate_all_intervention_ids_exist()
        print(dt.now(), "Validate frame_intervention_ids")
        validation_errors["frame_intervention_ids"] = self.validate_intervention_id_of_frames()

        return validation_errors

    def validate_config(self, cfg:dict):
        pass



from pydantic import BaseModel, NonNegativeFloat, NonNegativeInt, validator
from datetime import datetime
from typing import Optional
from pathlib import Path


class ImageMetadata(BaseModel):
    frame_number: Optional[int]
    path: Optional[Path]

    is_frame: bool
    is_extracted: bool

    class Config:
        allow_population_by_field_name = True
        # arbitrary_types_allowed = True
        json_encoders = {Path: str}
        schema_extra = {
            "example": {}
        }

    @validator("is_frame")
    def validate_frame_data(cls, v, values):
        if v:
            assert "frame_number" in values

    @validator("is_extracted")
    def validate_path_if_extracted(cls, v, values):
        if v:
            assert "path" in values


class InterventionMetadata(BaseModel):
    dicomaccessionnumberpseudo: Optional[str]
    dicomstydyinstanceuidpseudo: Optional[str]
    fps: Optional[NonNegativeFloat]
    duration: Optional[NonNegativeFloat]
    frames_total: Optional[NonNegativeInt]
    path: Optional[Path]
    is_video: bool
    intervention_date: Optional[datetime]
    intervention_type: str
    sap_case_id: Optional[int]
    sap_pat_id: Optional[int]
    st_export_id: Optional[int]

    class Config:
        allow_population_by_field_name = True
        # arbitrary_types_allowed = True
        json_encoders = {Path: str}
        schema_extra = {
            "example": {}
        }

    @validator("is_video")
    def validate_video_metadata(cls, v, values):
        if v:
            assert "path" in values
            assert "frames_total" in values
            assert "duration" in values
            assert "fps" in values
        return v

    def to_dict(self):
        r = self.dict()


from pathlib import Path
from ..ls.utils import *

class Labelstudio:
    """[summary]
    """

    def __init__(self, ls_url: str, ls_token: str, ls_config_path: Path, cfg: dict):
        self.url = ls_url
        self.token = ls_token
        self.ls_config_path = ls_config_path
        self.fileserver_prefix = cfg["url_endobox_extreme"] + ":" + cfg["port_fileserver"] + "/"


    def delete_project_by_id(self, project_id: int):
        delete_project_by_id(project_id, self.url, self.token)

    def generate_ls_project(self, label: str, description: str):
        return generate_ls_project(
            label,
            description,
            self.ls_config_path,
            self.url,
            self.token
        )

    def images_to_ls_project(self, images, label, description):
        new_project = self.generate_ls_project(label, description)
        new_project_id = new_project["id"]
        task_dump = [self.generate_ls_task(image, label, new_project_id) for image in images]
        task_dump = [_ for _ in task_dump if _]

        for task in task_dump:
            task = json.dumps(task)
            r = requests.post(
                f"{self.url}tasks/",
                headers={"Authorization": self.token, 'content-type': 'application/json'},
                data=task
                )
            assert r.status_code == 201

    def generate_ls_task(self, image, label, project_id):
        if "/extreme_storage/files/" in image["path"]:
            task = {
                "data": {
                    "image": image["path"].replace("/extreme_storage/files/", self.fileserver_prefix),
                    "prediction": str(image["predictions"][label]["value"]),
                    "_id": str(image["_id"])
                },
                "project": str(project_id)
            }
            return task


    def import_ls_project_annotations(self, project_id, db_images, import_as_validated:bool=False, verbose:bool = False):
        import_ls_project_annotations(
            project_id,
            self.url,
            self.token,
            db_images,
            import_as_validated,
            verbose
            )



from .fieldnames import *
from pathlib import Path
from collections import defaultdict
from typing import List

class FastCatAnnotation:
    """asd
    """
    def __init__(self, path: Path, cfg):
        self.skip_frame_factor = cfg["skip_frame_factor"]
        self.fast_cat_allowed_labels = cfg["fast_cat_allower_labels"]
        self.path = path
        self.flanks = self.read_fast_cat_annotation()
        self.labels = self.flanks_to_labels()
        assert path.exists()

    def read_fastcat_annotation(self, skip_frame_factor: int = None, allowed_label_list: List[str] = None):
        # Validate Metadata
        if not skip_frame_factor:
            skip_frame_factor = self.skip_frame_factor
        if not allowed_label_list:
            allowed_label_list = self.fast_cat_allowed_labels

        for label in self.annotation["metadata"]["classes"]:
            assert label in allowed_label_list

        label_list = [_.split(".")[0] for _ in allowed_label_list]
        label_list = list(set(label_list))

        flanks = {_: {"start": [], "stop": []} for _ in label_list}

        for superframe in self.annotation["superframes"]:
            for frame_annotation in superframe["childFrames"]:
                for label in frame_annotation["classes"]:
                    label, param = label.split(".")
                    flanks[label][param].append(frame_annotation["frameNumber"])
                
        for key, value in flanks.items():
            assert len(value["start"]) is len(value["stop"])
            flanks[key]["start"].sort()
            flanks[key]["stop"].sort()

            _matched = []
            for i, _start in enumerate(flanks[key]["start"]):
                _stop = flanks[key]["stop"][i]
                assert _stop >= _start
                _matched.append((_start, _stop))

            flanks[key] = _matched

        for key, value in flanks.items():
            for i, _flank in enumerate(value):
                if i == 0:
                    continue

                assert _flank[0] > value[i - 1][1]


        flanks["skip_frame_factor"] = skip_frame_factor

        return flanks

    def flanks_to_labels(self, flanks_dict: dict = None):
        if not flanks_dict:
            flanks_dict = self.flanks
        frame_labels = defaultdict(dict)
        skip_frame_factor = flanks_dict["skip_frame_factor"]
        for label, _flanks in flanks_dict.items():
            if label == "skip_frame_factor":
                continue
            for _flank in _flanks:
                for n_frame in range(_flank[0], _flank[1]+1):
                    frame_labels[n_frame*skip_frame_factor][label] = True

        return frame_labels



    


from pydantic import (
    BaseModel,
    NonNegativeInt,
    validator
)
from typing import List, Any
from datetime import datetime
from .base import Flank

class Annotation(BaseModel):
    source: str #constr(regex=r"")
    annotator_id: NonNegativeInt
    date: datetime
    name: str

    @validator("date")
    def rm_ms(cls, v):
        v = v.replace(microsecond=0)
        return v

class BinaryAnnotation(Annotation):
    value: bool
    choices: List[bool]
    label_type = "binary"

class MultilabelAnnotation(Annotation):
    value: NonNegativeInt
    choices: List[str]
    label_type = "multilabel"

class MultichoiceAnnotation(Annotation):
    value: List[NonNegativeInt]
    choices: List[str]
    label_type = "multichoice"

class VideoSegmentationAnnotation(Annotation):
    value: List[Flank]
    label_type = "video_segmentation"

class PolypReportAnnotation(Annotation):#######################
    value: Any



from typing import Dict, List, Optional, Union
from bson import ObjectId
from pydantic import BaseModel, Field, validator
from pathlib import Path
from ..media.metadata import get_video_meta
from ..extern.base import path_to_timestamp,video_extern_exists_intern, video_extern_get_intern_id
from .base import Flank
from datetime import datetime, tzinfo

INTERVENTION_TYPE_MAPPING = {
    "Coloscopy": "Koloskopie",
    "Unknown": "Unknown",
    "Gastroscopy": "Gastroskopie"
}

class VideoExtern(BaseModel):
    id_extern: int = Field(alias="videoId")
    path: Path = Field(alias= "videoPath")
    images_path: Optional[Path] = Field(alias= "imagesPath")
    annotation_paths: Optional[Path]
    video_remark: Optional[str] #= Field(alias="")
    intervention_histo_text: Optional[str] =  Field(alias = "patho")
    intervention_report_text: Optional[str] = Field(alias = "report")
    origin: Optional[str] = Field(alias = "center")
    intervention_type: str = Field(alias="videoType")
    annotated: bool


    class Config:
        # allow_population_by_field_name = True
        json_encoders = {Path: str}
        schema_extra = {"example": {}}

    @validator('intervention_histo_text')
    def rm_whitespace_histo(cls, v) -> str:
        if v:
            v = v.replace("\n", " ")
            v = v.replace("  ", " ")
            v.strip()
            return v

    @validator('intervention_report_text')
    def rm_whitespace_report(cls, v) -> str:
        if v:
            v = v.replace("\n", " ")
            v = v.replace("  ", " ")
            v.strip()
            return v

    def get_video_meta(self):
        return get_video_meta(self.path)

    def map_intervention_type(self):
        _type=self.intervention_type
        assert _type in INTERVENTION_TYPE_MAPPING
        _type = INTERVENTION_TYPE_MAPPING[_type]
        return _type

    def exists_intern(self, db_interventions):
        return video_extern_exists_intern(self.id_extern, db_interventions)

    def get_id_intern(self, db_interventions):
        _id = video_extern_get_intern_id(self.id_extern, db_interventions)
        if _id: return _id

    def to_intervention_dict(self, db_interventions, time_format = "%Y-%m-%d_%H-%M-%S"):
        intervention_dict = self.dict()
        id_intern = self.get_id_intern(db_interventions)
        if id_intern:
            intervention_dict["id"] = id_intern
        intervention_dict["frames"] = {}
        intervention_dict["freezes"] = {}
        intervention_dict["image_annotations_sum"] = {}
        intervention_dict["image_predictions_sum"] = {}
        intervention_dict["video_key"] = self.path.name
        if self.intervention_histo_text: 
            intervention_dict["intervention_histo_text"] = {"text": self.intervention_histo_text}
        if self.intervention_report_text:
            intervention_dict["intervention_report_text"] = {"text": self.intervention_report_text}


        metadata_dict = self.get_video_meta()
        metadata_dict["path"] = self.path
        metadata_dict["is_video"] = True
        metadata_dict["intervention_date"] = path_to_timestamp(self.path, time_format, self.origin)
        metadata_dict["intervention_type"] = self.map_intervention_type()

        intervention_dict["metadata"] = metadata_dict

        return intervention_dict

class ExternFlank(BaseModel):
    name: str = Field(alias = "annotationName")
    value: Union[bool, str] = Field(alias = "annotationValue")
    start: int = Field(alias = "startFrame")
    stop: int = Field(alias = "endFrame")

    @validator("name")
    def name_to_lower(cls, v):
        return v.lower()

class ExternVideoFlankAnnotation(BaseModel):
    extern_annotator_id: int = Field(alias = "userId")
    id_extern: int = Field(alias = "videoId")
    video_key: str = Field(alias = "videoName")
    date: datetime = Field(alias = "sessionDate")
    label_group_id: int = Field(alias = "labelGroupId")
    label_group_name: str = Field(alias = "labelGroupName")
    flanks: List[ExternFlank] = Field(alias = "sequences")

    @validator("date")
    def rm_tz(cls, v):
        v = v.replace(microsecond=0)
        return v

class ExternAnnotatedVideo(BaseModel):
    id_extern: int = Field (alias = "videoId")
    video_key: str = Field(alias = "videoName")
    date: datetime = Field(alias = "lastAnnotationSession")

    @validator("date")
    def rm_tz(cls, v):
        v = v.replace(tzinfo=None)
        v = v.replace(microsecond=0)
        return v



from typing import Dict, List, Optional
from bson import ObjectId
from pydantic import BaseModel, Field, NonNegativeInt, validator
import pandas as pd

from .annotation import PolypReportAnnotation, VideoSegmentationAnnotation
from .base import PyObjectId
from .prediction import VideoSegmentationPrediction
from .text import Text, Token
from .metadata import InterventionMetadata
from pathlib import Path

class Intervention(BaseModel):
    id: Optional[PyObjectId] = Field(alias="_id")
    video_key: Optional[str]
    origin: str
    id_extern: Optional[int]
    age: Optional[int]
    gender: Optional[int]
    frames: Dict[int, PyObjectId]
    freezes: Dict[int, PyObjectId]
    image_annotations_sum: Dict[str, Dict[str, NonNegativeInt]]
    image_predictions_sum: Dict[str, Dict[str, NonNegativeInt]]
    video_segments_annotation: Optional[Dict[str, VideoSegmentationAnnotation]]
    video_segments_prediction: Optional[Dict[str, VideoSegmentationPrediction]]
    intervention_report_text: Optional[Text]
    intervention_histo_text: Optional[Text]
    intervention_report_structured: Optional[Dict[str, str]]
    intervention_report_id: Optional[int]
    intervention_histology_structured: Optional[Dict[str, str]]
    text_tokens: Optional[Dict[str, List[Token]]]
    polyp_report_annotations: Optional[List[PolypReportAnnotation]]
    metadata: InterventionMetadata

    class Config:
        allow_population_by_field_name = True
        # arbitrary_types_allowed = True
        json_encoders = {ObjectId: str, InterventionMetadata: dict, Path: str}
        schema_extra = {"example": {}}

    # @validator("age")
    # def age_is_valid(cls, v):
    #     assert v >= 0
    #     return v

    @validator("metadata")
    def validate_fields_by_metadata(cls, v, values):
        if v.is_video:
            assert "video_segments_annotation" in values
            assert "video_segments_prediction" in values
            assert "video_key" in values

        return v

    def to_dict(self):
        r = self.dict(by_alias=True, exclude_none = True)
        if "path" in r["metadata"]:
            r["metadata"]["path"] = str(r["metadata"]["path"])
        r["frames"] = {str(n): _id for n, _id in r["frames"].items()}
        r["freezes"] = {str(n): _id for n, _id in r["freezes"].items()}

        return r

    def create_in_db(self, db_interventions):
        if self.video_key:
            _ = db_interventions.find_one({"video_key": self.video_key})
            assert not _

    def merge_id_extern(self, db_interventions):
        r = db_interventions.update_one({"video_key": self.video_key}, {"$set": {"id_extern": self.id_extern}})
        return r

    def summary(self):
        summary = self.dict(exclude={"frames", "freezes", "text_tokens"})
        summary["n_db_frames"] = len(self.frames)
        summary["n_db_freezes"] = len(self.freezes)
        return summary

    def frame_df(self):
        df = pd.DataFrame.from_records([self.frames]).transpose().reset_index()
        df.columns = ["frame_number", "id"]
        df = df.sort_values("frame_number")
        return df


import warnings
from bson.objectid import ObjectId
from pymongo.collection import Collection
import pandas as pd
from typing import List
from pathlib import Path
from .fieldnames import *
from .db_images import DbImages
from collections import defaultdict
from .db_interventions import DbInterventions


class DbTests:
    """[summary]
    """
    def __init__(self, db_tests: Collection, db_images: DbImages, db_interventions: DbInterventions, cfg: dict):
        self.db = db_tests
        self.db_images = db_images
        self.db_interventions = db_interventions
        self.cfg = cfg


    # CRUD
    ## Create / Update
    def read_test_data(self, test_data: List[dict]):
        db_test_data_ids = defaultdict(list)
        db_test_data_keys = defaultdict(list)

        for test_data_dict in test_data:
            path = Path(test_data_dict["path"])
            label_list = test_data_dict["label_list"]

            if path.suffix == ".csv":
                df = pd.read_csv(path)
            elif path.suffix == ".xlsx":
                df = pd.read_excel(path)
            video_keys = df[FIELDNAME_VIDEO_KEY].dropna().to_list()
            test_interventions = self.db_interventions.db.find({FIELDNAME_VIDEO_KEY: {"$in": video_keys}},{"_id": 1})
            intervention_ids = [_["_id"] for _ in test_interventions]

            for label in label_list:
                db_test_data_ids[label].extend(intervention_ids)
                db_test_data_keys[label].extend(video_keys)

        for label in db_test_data_ids.keys():
            self.db.update_one(
                {"label": label},
                {
                    "$set": {
                        "intervention_ids": db_test_data_ids[label],
                        "video_keys": db_test_data_keys[label]
                    }
                    },
                upsert = True)


    ## Read
    def get_test_data_interventions(self, label: str) -> List[ObjectId]:
        test_set = self.db.find_one({"label": label})
        if test_set:
            return test_set["intervention_ids"]

        else:
            warnings.warn(f"No Test Set for Label {label} found")
            return []




from ..classes.intervention import Intervention
from ..classes.image import Image
from ..classes.train_data import TrainData, TrainDataDb

def get_intervention(identifier, db_interventions, id_is_key = True):
    if id_is_key: param = "video_key"
    else: param = "_id"

    return Intervention(**db_interventions.find_one({param: identifier}))

def get_images_by_id_list(id_list, db_images):
    r = db_images.find({"_id": {"$in": id_list}})
    image_list = [Image(**_) for _ in r]

    return image_list

def get_train_data(name, db_train_data):
    train_data_db = TrainDataDb(**db_train_data.find_one({"name": name}))
    _dict = train_data_db.to_dict_intern()
    train_data = TrainData(**_dict)

    return train_data


from .multilabel_googlenet import MultilabelGoogleNet
from ..classes.config import AiLabelConfig
from pathlib import Path

LOOKUP_MODELS = {
    "multilabel_googlenet": MultilabelGoogleNet
}

def get_model(ai_config:AiLabelConfig, ckpt_path:Path, eval = True, cuda = True):
    model = LOOKUP_MODELS[ai_config.ai_settings.base_model_name]
    model = MultilabelGoogleNet.load_from_checkpoint(ckpt_path.as_posix())
    if cuda:
        model.to(0)
    if eval:
        model.eval()
        model.freeze()

    return model


import subprocess
from pathlib import Path
import json
import warnings
import cv2

def get_video_meta(path: Path) -> dict:
    command = f"ffprobe -hide_banner -loglevel fatal \
        -show_error -show_format -show_streams -show_chapters \
            -show_private_data -print_format json '{path.as_posix()}'"

    _ = subprocess.run(command, capture_output=True, shell = True)

    meta = json.loads(_.stdout)
    # assert len(meta["streams"]) == 1
    try:
        stream = meta["streams"][0]
        fps_strings = stream["r_frame_rate"].split("/")
        fps = float(fps_strings[0])/float(fps_strings[1])
        duration = float(meta["format"]["duration"])
        frames_total = fps*duration
    
    except:
        warnings.warn(f"Could not Generate Metadata for {path}, using cv2")
        cap = cv2.VideoCapture(path.as_posix())
        fps = cap.get(cv2.CAP_PROP_FPS)

        frames_total = 0
        success, frame = cap.read()

        while success:
            frames_total +=1
            success, frame = cap.read()

        duration = frames_total / fps

    meta = {
        "fps": fps,
        "frames_total": frames_total,
        "duration": duration
    }

    return meta


from typing import Dict, Union, Optional
from bson import ObjectId
from pydantic import BaseModel, Field, NonNegativeInt, validator

from .annotation import BinaryAnnotation, MultilabelAnnotation, MultichoiceAnnotation
from .base import PyObjectId
from .prediction import BinaryPrediction, MultilabelPrediction, MultichoicePrediction
from .metadata import ImageMetadata
from .text import Text

class Image(BaseModel):
    id: Optional[PyObjectId] = Field(alias="_id")
    intervention_id: PyObjectId = Field(default_factory=PyObjectId)
    origin: str
    video_key: Optional[str]
    predictions: Dict[str, Union[BinaryPrediction, MultilabelPrediction, MultichoicePrediction]]
    predictions_smooth: Optional[Dict[str, Union[BinaryPrediction, MultilabelPrediction, MultichoicePrediction]]]
    annotations: Dict[str, Union[BinaryAnnotation, MultilabelAnnotation, MultichoiceAnnotation]]
    image_caption: Optional[Text]
    metadata: ImageMetadata

    @validator("metadata")
    def validate_metadata(cls, v, values):
        if v.is_frame:
            assert "video_key" in values
        if v.is_extracted:
            assert v.path != None
        return v

    class Config:
        allow_population_by_field_name = True
        json_encoders = {ObjectId: str, ImageMetadata: dict}
        schema_extra = {
            "example": {}
        }

    def to_dict(self):
        r = self.dict(by_alias=True, exclude_none = True)
        if "path" in r["metadata"]:
            r["metadata"]["path"] = str(r["metadata"]["path"])
        return r



from sklearn.metrics import classification_report

def get_classification_report(results_df, prediction_type):
    s1 = results_df._type == "annotations"
    y_true = results_df[s1].value.to_list()
    s1 = results_df._type == prediction_type
    y_pred = results_df[s1].value.to_list()
    _report = classification_report(y_true, y_pred, output_dict = True)
    return _report


from collections import defaultdict

def annotations_by_label_group(video_annotations):
    '''
    Returns dict with key = label group and list of annotation objects.
    List is ordered by date, index 0 is most recent
    '''
    
    annotation_dict = defaultdict(list)
    for annotation in video_annotations:
        annotation_dict[annotation.label_group_id].append(annotation)

    for group in annotation_dict.keys():
        annotation_dict[group].sort(key= lambda x: x.date, reverse = True)

    return annotation_dict


def get_labels_with_test_annotation(group_ids, test_label_group_dict):
    labels_with_test_annotation = {}
    for _id in group_ids:
        if _id in test_label_group_dict:
            for key, value in test_label_group_dict[_id].items():
                labels_with_test_annotation[key] = value

    return labels_with_test_annotation


def get_flank_dict(video_annotation_dict, test_label_group_dict):
    group_ids = list(video_annotation_dict.keys())
    lookup = get_labels_with_test_annotation(group_ids, test_label_group_dict)
    flanks = {"test": [], "train": []}
    for _id, annotations in video_annotation_dict.items():
        # annotations is a list ordered by date, starting with most recent
        annotation = annotations[0]
        if _id in test_label_group_dict:
            if _id == 4:
                for annotation in annotations:
                    flanks["test"].extend(annotation.flanks)
            else: 
                flanks["test"].extend(annotation.flanks)
        else:
            flanks["train"].extend([_ for _ in annotation.flanks if _.name not in lookup])
        
    return flanks, lookup



from pymongo import collection
from typing import List, Dict
from bson import ObjectId
from ..fieldnames import *
from collections import defaultdict
from ..classes.intervention import Intervention
from .get_objects import get_intervention
from pathlib import Path

def get_extern_ids(collection):
    _ = collection.aggregate([
        {"$match": {"id_extern": {"$exists": True}}},
        {"$project": {"id_extern": 1}}
    ])
    _ = [_["id_extern"] for _ in _]
    return _

def video_keys(collection):
    _ = collection.aggregate([
        {"$match": {"video_key": {"$exists": True}}},
        {"$project": {"video_key": 1}}
    ])
    _ = [_["id_extern"] for _ in _]
    return _

def get_all_in_batches(collection: collection, limit: int = 10000, aggregation: List = [{"$match": {}}], last_id: ObjectId = None):
    aggregation = aggregation.copy()
    if last_id:
        aggregation.append({"_id": {"$gt": last_id}})
    aggregation.append({"$limit": limit})
    cursor = collection.aggregate(aggregation)

    return cursor


def filter_images_by_frame_diff(img_list: List[dict], min_frame_diff: int = 10):
    # Holy Moly, Bumms ineffizient, muss nachgebessert werden
    filtered_images_dict = defaultdict(list)

    for img in img_list:
        intervention_id = img[IMG_INTERVENTION_ID]

        if not img[METADATA][IMG_IS_FRAME]:
            filtered_images_dict[intervention_id].append(img)
            continue

        append = True
        n_frame = img[METADATA][IMG_FRAME_NUMBER]
        
        if filtered_images_dict[intervention_id]:
            for _img in filtered_images_dict[intervention_id]:
                frame_diff = abs(_img[METADATA][IMG_FRAME_NUMBER] - n_frame)
                if frame_diff < min_frame_diff:
                    append = False
                    break
        
        if append:
            filtered_images_dict[intervention_id].append(img)

    filtered_images = []
    for key, value in filtered_images_dict.items():
        filtered_images.extend(value)
    
    return filtered_images

def filter_existing_frames_by_intervention(video_key, frame_paths, db_interventions) -> Dict[int, Path]:
    intervention = get_intervention(video_key, db_interventions)
    image_paths = {n: path for n, path in frame_paths.items() if n not in intervention.frames}
    return image_paths

def get_base_image_dict_from_intervention(intervention:Intervention):
    base_image_dict = {
        "intervention_id": intervention.id,
        "origin": intervention.origin,
        "video_key": intervention.video_key,
        "predictions": {},
        "annotations": {},
        "metadata": {}
    }
    return base_image_dict


import requests
from ..classes.extern import ExternAnnotatedVideo, ExternVideoFlankAnnotation
from urllib3.exceptions import InsecureRequestWarning
from typing import List
from ..classes.extern import VideoExtern

requests.packages.urllib3.disable_warnings(category=InsecureRequestWarning)


def get_extern_annotations(url, auth):
    r = requests.get(f"{url}/GetVideosWithAnnotations", auth = auth, verify = False)
    assert r.status_code == 200
    r = [ExternAnnotatedVideo(**_) for _ in r.json()]

    return r

def get_extern_video_annotation(video_key, url, auth):
    r = requests.get(url+"/GetAnnotationsByVideoName/"+video_key, auth = auth, verify = False)
    assert r.status_code == 200
    annotations = [ExternVideoFlankAnnotation(**_) for _ in r.json()]

    return annotations


def extract_video_intervals(_json, url, auth):
    r = requests.post(url+"/PostVideoToImageIntervals", auth = auth,json=_json, verify=False)
    return r

def get_extern_interventions(url, auth) -> List[dict]:
    r = requests.get(f"{url}/GetVideosExtern", auth=auth, verify = False)
    assert r.status_code == 200
    videos = [VideoExtern(**_) for _ in r.json()]

    return videos


from typing import Optional
from typing import Tuple
from ..classes.train_data import TrainData

from pytorch_lightning import LightningDataModule
from torch.utils.data import DataLoader, WeightedRandomSampler
from torch.utils.data import Dataset
from torch.utils.data import random_split
import torch

from sklearn.model_selection import StratifiedShuffleSplit
import numpy as np

from ..datasets.binary_image_classification_ds import BinaryImageClassificationDS


class MultilabelImageClassificationDM(LightningDataModule):

    """
    Example of LightningDataModule for MNIST dataset.

    A DataModule implements 5 key methods:
        - prepare_data (things to do on 1 GPU/TPU, not on every GPU/TPU in distributed mode)
        - setup (things to do on every accelerator in distributed mode)
        - train_dataloader (the training dataloader)
        - val_dataloader (the validation dataloader(s))
        - test_dataloader (the test dataloader(s))

    This allows you to share a full dataset without explaining how to download,
    split, transform and process the data

    Read the docs:
        https://pytorch-lightning.readthedocs.io/en/latest/extensions/datamodules.html
    """

    def __init__(
        self,
        train_data: TrainData,
        scaling: int,
        test_size: Tuple[float, float, float] = 0.2,
        batch_size: int = 64,
        num_workers: int = 0,
        pin_memory: bool = False,
        **kwargs,
    ):
        super().__init__()

        self.train_data_object = train_data

        # self.paths = np.array([_.as_posix() for _ in train_data.paths])
        # self.labels = np.array(train_data.labels)
        self.scaling = scaling
        self.test_size = test_size
        self.batch_size = batch_size
        self.num_workers = num_workers
        self.pin_memory = pin_memory
        self.num_classes = train_data.n_classes()

        self.data_train: Optional[Dataset] = None
        self.data_val: Optional[Dataset] = None
        self.data_test: Optional[Dataset] = None

    def prepare_data(self):
        """Download data if needed. This method is called only from a single GPU.
        Do not use it to assign state (self.x = y)."""
        pass

    def setup(self, stage: Optional[str] = None):
        """Load data. Set variables: self.data_train, self.data_val, self.data_test."""

        # sss = StratifiedShuffleSplit(n_splits = 1, test_size = 0.1, random_state = 42)
        # train_indexes, test_indexes = [_ for _ in sss.split(
        #     self.train_data_object.paths,
        #     self.train_data_object.labels
        # )][0]

        # x_train, x_test, y_train, y_test = self.train_data_object.group_strat_split(
        #     test_size=self.test_size)
        labels = np.array(self.train_data_object.labels)
        paths = np.array([_.as_posix() for _ in self.train_data_object.paths])
        is_val = np.array(self.train_data_object.is_val)

        x_train = paths[~is_val]
        x_test = paths[is_val]
        y_train = labels[~is_val]
        y_test = labels[is_val]


        class_sample_count = np.array(
            [len(np.where(y_train == t)[0]) for t in np.unique(y_train)])

        weight = 1. / class_sample_count
        samples_weight = np.array([weight[t] for t in y_train])
        samples_weight = torch.from_numpy(samples_weight)
        samples_weight = samples_weight.double()
        self.train_sampler = WeightedRandomSampler(samples_weight, len(samples_weight))

        y_train = torch.from_numpy(y_train).long()

        self.train_dataset = BinaryImageClassificationDS(
            x_train, y_train, self.scaling
        )

        self.test_dataset = BinaryImageClassificationDS(
            x_test, y_test, self.scaling
        )

    def train_dataloader(self):
        return DataLoader(
            dataset=self.train_dataset,
            batch_size=self.batch_size,
            num_workers=self.num_workers,
            pin_memory=self.pin_memory,
            sampler = self.train_sampler
        )

    def val_dataloader(self):
        return DataLoader(
            dataset=self.test_dataset,
            batch_size=self.batch_size,
            num_workers=self.num_workers,
            pin_memory=self.pin_memory,
            shuffle=False,
        )

    # def test_dataloader(self):
    #     return DataLoader(
    #         dataset=self.data_test,
    #         batch_size=self.batch_size,
    #         num_workers=self.num_workers,
    #         pin_memory=self.pin_memory,
    #         shuffle=False,
    #     )



import warnings
import pandas as pd
from pathlib import Path
from collections import defaultdict
from typing import Any, List
from bson.objectid import ObjectId
from pymongo.collection import Collection
from .fieldnames import *
import os
from datetime import datetime as dt
import numpy as np
import re
import subprocess
import json
import cv2
from operator import itemgetter

def generate_default_dict():
    return defaultdict(dict)

def generate_list_default_dict():
    return defaultdict(list)

def generate_frame_path(video_key:str, base_path_frames: Path) -> Path:
    return base_path_frames.joinpath(video_key)

# Crud
def match_logical_and_aggregation(condition_list: List[str]):
    agg_element = {
        "$match": {
            "$and": [
                _ for _ in condition_list
            ]
        }
    }

    return agg_element

def logical_or_aggregation(condition_list):
    agg_element = {
        "$or": [_ for _ in condition_list]
    }

    return agg_element

def field_exists_query(fieldname: str, exists: bool = True) -> dict:
    return {
        fieldname: {"$exists": exists}
    }

def field_value_query(fieldname: str, value: Any) -> dict:
    return {
        fieldname: value
        }

def fieldvalue_in_list_query(fieldname: str, values: List[Any]) -> dict:
    return {
        fieldname: {"$in": values}
    }

def fieldvalue_nin_list_query(fieldname: str, values: List[Any]) -> dict:
    return {
        fieldname: {"$nin": values}
    }

def get_intervention_db_template(**kwargs):
    
    template = {
        FIELDNAME_EXTERNAL_ID: None,
        FIELDNAME_FRAMES: {},
        FIELDNAME_FREEZES: {},
        FIELDNAME_VIDEO_KEY: None,
        FIELDNAME_VIDEO_PATH: None,
        FIELDNAME_INTERVENTION_TYPE: None,
        FIELDNAME_TOKENS: {
            FIELDNAME_TOKENS_REPORT: [],
            FIELDNAME_TOKENS_PATHO: []
        }
    }

    # if "intervention_dict" in kwargs:
    #     for key, value in kwargs["intervention_dict"]:
    #         if key in template:
    #             template[key]=value

    return template


def get_image_db_template(
    origin: str = None,
    intervention_id: ObjectId = None,
    path: str = None,
    n_frame: int = None,
    image_type: str = None
):
    """
    """

    template = {
        FIELDNAME_ORIGIN: origin,
        FIELDNAME_INTERVENTION_ID: intervention_id,
        FIELDNAME_IMAGE_PATH: path,
        FIELDNAME_FRAME_NUMBER: n_frame,
        FIELDNAME_IMAGE_TYPE: image_type,
        FIELDNAME_LABELS: {},
        FIELDNAME_PREDICTIONS: {},

    }
    return template


def delete_frame_from_intervention(
    frame_dict: dict,
    db_interventions: Collection
    ):
    n_frame = frame_dict[FIELDNAME_FRAME_NUMBER]
    frame_id = frame_dict["_id"]
    intervention_id = frame_dict[FIELDNAME_INTERVENTION_ID]
    frame_path = Path(frame_dict[FIELDNAME_IMAGE_PATH])
    assert frame_path.exists()
    

    intervention = db_interventions.find_one({"_id": intervention_id})
    assert intervention
    assert "frames" in intervention
    assert str(n_frame) in intervention[FIELDNAME_FRAMES]
    assert frame_id is intervention[FIELDNAME_FRAMES]
    db_interventions.update_one(
        {"_id": intervention_id},
        {"$unset": {f"{FIELDNAME_FRAMES}.{n_frame}": ""}}
        )

    os.remove(frame_path)

    return True


def datetime_from_video_key(video_key, re_pattern, dt_pattern):
    date_string = re.search(re_pattern, video_key)
    if date_string:
        date_string = date_string.group()
        intervention_date = dt.strptime(date_string, dt_pattern)

        return intervention_date

def get_video_meta(path: Path) -> dict:
    command = f"ffprobe -hide_banner -loglevel fatal \
        -show_error -show_format -show_streams -show_chapters \
            -show_private_data -print_format json '{path.as_posix()}'"

    _ = subprocess.run(command, capture_output=True, shell = True)

    meta = json.loads(_.stdout)
    # assert len(meta["streams"]) == 1
    try:
        stream = meta["streams"][0]
        fps_strings = stream["r_frame_rate"].split("/")
        fps = float(fps_strings[0])/float(fps_strings[1])
        duration = float(meta["format"]["duration"])
        frames_total = fps*duration
    
    except:
        warnings.warn(f"Could not Generate Metadata for {path}, using cv2")
        cap = cv2.VideoCapture(path.as_posix())
        fps = cap.get(cv2.CAP_PROP_FPS)

        frames_total = 0
        success, frame = cap.read()

        while success:
            frames_total +=1
            success, frame = cap.read()

        duration = frames_total / fps

    meta = {
        FIELDNAME_FPS: fps,
        FIELDNAME_FRAMES_TOTAL: frames_total,
        FIELDNAME_VIDEO_DURATION: duration
    }

    return meta

def get_frames_total(cap: cv2.VideoCapture) -> int:
    frames_total = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))-2
    assert frames_total
    return frames_total


def get_fps(cap: cv2.VideoCapture) -> float:
    fps = cap.get(cv2.CAP_PROP_FPS)
    assert fps
    return fps

def prediction_df_long_to_wide(df_long: pd.DataFrame) -> pd.DataFrame:
    return df_long.pivot_table(
        index=[FIELDNAME_FRAME_NUMBER, "_id"],
        columns=[COLNAME_AI_NAME],
        values=[FIELDNAME_PREDICTION_LABEL, FIELDNAME_PREDICTION_VALUE]
        )

def parse_stats_dict_name(name, value):
    keys = name.split(".")
    record = {}
    record[DF_COL_ENTITY] = keys[0]
    record[DF_COL_VALUE_TYPE] = keys[1]
    record[DF_COL_DATE] = dt.now()
    record[DF_COL_VALUE] = value
    
    new_attribute = []
    for i, key in enumerate(keys[2:]):
        new_attribute.append(key)
        if len(new_attribute) == 2:
            record.update({
                new_attribute[0]: new_attribute[1]
            })
            new_attribute = []
    return record

def get_count(db_images, entity, match_conditions_dict) -> dict:
    count = {}
    prefixes = [entity, PREFIX_COUNT]
    match_conditions = {}

    for prefix, match_condition in match_conditions_dict.items():
        prefixes.append(prefix)
        match_conditions.update(match_condition)

    prefix = ".".join(prefixes)
    count[prefix] = db_images.count_documents(match_conditions)
    return count


def get_grouped_count(db_images, entity, fieldname, additional_match_conditions: dict = None) -> dict:
    match_conditions = {fieldname: {"$nin": ["", None, [], {}]}}
    prefixes = [entity, PREFIX_COUNT]

    if additional_match_conditions:
        for prefix, _match_conditions in additional_match_conditions.items():
            prefixes.append(prefix)
            match_conditions.update(_match_conditions)

    _grouped_count = db_images.aggregate([
        {
            "$match": match_conditions
        },
        {
            "$group": {
                "_id": "$"+fieldname,
                PREFIX_COUNT: {"$sum": 1} 
            }
        }
    ])

    prefix = ".".join(prefixes)

    grouped_count = {
        f"{prefix}.{fieldname}.{_['_id']}": _[PREFIX_COUNT] for _ in _grouped_count
    }

    return grouped_count


def get_grouped_dict_count(db_images, entity, fieldname, additional_match_conditions: dict = None, additional_group_conditions: dict = None):
    match_conditions = {fieldname: {"$nin": ["", None, [], {}]}}
    prefixes = [entity, PREFIX_COUNT]

    if additional_match_conditions:
            for prefix, _match_conditions in additional_match_conditions.items():
                prefixes.append(prefix)
                match_conditions.update(_match_conditions)

    _group_id = {
        "key": "$target.k",
        "value": "$target.v"
    }

    _project_conditions = {
        "target": {"$objectToArray": "$" + fieldname}
    }

    if additional_group_conditions:
        _group_id.update(additional_group_conditions)
        _project_conditions.update(additional_group_conditions)

    aggregation = [
        {
            "$match": match_conditions
        },
        {
            "$project": _project_conditions
        },
        {"$unwind": "$target"},
        {
            "$group": {
                "_id": _group_id,
                "count": {"$sum": 1}
            }
        }
    ]

    results = [_ for _ in db_images.aggregate(aggregation)]

    grouped_count = {}
    for result in results:
        _id_values = result["_id"]
        target_keys = []
        _prefixes = prefixes.copy()
        _prefixes.extend([result["_id"]["key"],result["_id"]["value"]])
        # prefix = ".".join(_prefixes)
        for key, value in _id_values.items():
            if key == "key" or key == "value":
                continue
            _prefixes.extend([str(key), str(value)])
        _prefixes = [str(_) for _ in _prefixes]
        dict_key = ".".join(_prefixes)
        _count = result["count"]

        grouped_count[dict_key] = _count


    return grouped_count

def calculate_stats_dict(db, stats_queries, return_records: bool = True):
    stats_dict = {}
    entity = stats_queries["entity"]
    stats_dict[f"{entity}.{PREFIX_COUNT}"] = db.count_documents({})
    for query in stats_queries["get_count"]:
        stats_dict.update(get_count(db, entity, query))
    for query in stats_queries["get_grouped_count"]:
        stats_dict.update(get_grouped_count(db, entity, **query))
    for query in stats_queries["get_grouped_dict_count"]:
        stats_dict.update(get_grouped_dict_count(db, entity, **query))

    
    if return_records:
        records = []
        for name, value in stats_dict.items():
            record = parse_stats_dict_name(name, value)
            records.append(record)
        return records
        

    else:
        return stats_dict


def get_intervention_image_predictions(intervention_id, db_images):
    aggregation = [
        {
            "$match": {
                FIELDNAME_INTERVENTION_ID: intervention_id
            }
        },
        {
            "$project": {
                "predictions": {"$objectToArray": "$predictions"}
            }
        },
        {
            "$unwind": "$predictions"
        },
        {
            "$project": {
                "label": "$predictions.k",
                "value": "$predictions.v.label"
            }
        },
        {
            "$group": {
                "_id": {
                    "label": "$label",
                    "value": "$value"
                },
                "count": {"$sum": 1}
            }
        }
    ]

    _predictions = db_images.aggregate(aggregation)
    predictions = defaultdict(generate_default_dict)
    for _ in _predictions:
        predictions[str(_["_id"]["label"])][str(_["_id"]["value"])] = _["count"]

    return predictions

def get_intervention_image_annotations(intervention_id, db_images):
    additional_match_conditions = {
        f"{FIELDNAME_INTERVENTION_ID}.{str(intervention_id)}": {FIELDNAME_INTERVENTION_ID: intervention_id}
    }

    _label_count = get_grouped_dict_count(db_images, PREFIX_IMAGE, FIELDNAME_LABELS, additional_match_conditions)

    label_count = defaultdict(generate_default_dict)

    for key, value in _label_count.items():
        components = key.split(".")
        label_count[components[-2]][components[-1]]=value

    return label_count


def get_if_in_dict(_dict, fieldnames):
    value = _dict

    for fieldname in fieldnames:
        if fieldname in value:
            value = value[fieldname]
        else: 
            value = None
            break

    return value


def smooth_predictions(df, label, n_smooth, use_future: bool = True):
    select_1 = df[COLNAME_RESULT_NAME] == label
    select_2 = df[COLNAME_RESULT_TYPE] == RESULT_TYPE_PREDICTION
    select = select_1 & select_2

    selection = df.loc[select].sort_values(FIELDNAME_FRAME_NUMBER, ascending = True)
    consecutive_df_list = split_df_to_consecutive_frame_dfs(selection)
    convolution_result = np.empty(0, dtype = np.bool)
    ids = []
    frame_numbers = []

    for _selection in consecutive_df_list:
        if not len(_selection)>n_smooth:
            _frame_numbers = _selection[FIELDNAME_FRAME_NUMBER].to_list()
            _ids = _selection["_id"].to_list()
            _convolution_result = np.empty((len(_selection)))
            _convolution_result[:] = np.nan
        else:
            prediction_array = np.array(_selection[COLNAME_RESULT_LABEL])
            _frame_numbers = _selection[FIELDNAME_FRAME_NUMBER].to_list()
            _ids = _selection["_id"].to_list()
            if use_future:
                convolution_filter = np.ones(n_smooth)/n_smooth
            else: 
                convolution_filter = np.ones(n_smooth)/(int(n_smooth/2))
                convolution_filter[int(n_smooth/2)+1:] = 0
                
            _convolution_result = np.convolve(prediction_array, convolution_filter, mode = "same")

        convolution_result = np.append(convolution_result, _convolution_result)
        ids.extend(_ids)
        frame_numbers.extend(_frame_numbers)

    return convolution_result, frame_numbers, ids

def values_to_bool(values, threshold):
    values[values>threshold] = 1
    values[values<=threshold] = 0
    values = values.astype(bool)
    return values

def get_smooth_prediction_df(df, label, n_smooth, threshold, use_future: bool = True):
    values_smooth, frame_numbers, ids = smooth_predictions(df, label, n_smooth)
    labels_smooth = values_to_bool(values_smooth, threshold)

    records = []

    for n in range(len(values_smooth)):
        records.append({
            "_id": ids[n],
            FIELDNAME_FRAME_NUMBER: frame_numbers[n],
            COLNAME_PREDICTION_VALUE: values_smooth[n],
            COLNAME_RESULT_LABEL: labels_smooth[n],
            COLNAME_RESULT_NAME: label,
            COLNAME_RESULT_TYPE: RESULT_TYPE_PREDICTION_SMOOTH
        })

    df = pd.DataFrame.from_records(records)

    return df

def get_prediction_flanks(intervention, instance_min_seconds: int = 1):
    from tqdm import tqdm
    has_framecount = False
    frames_total = None
    fps = None
    _flanks = defaultdict(list)
    flanks = defaultdict(dict)

    has_meta = FIELDNAME_VIDEO_METADATA in intervention.intervention
    if has_meta:
        has_framecount = FIELDNAME_FRAMES_TOTAL in intervention.intervention[FIELDNAME_VIDEO_METADATA]

    if has_framecount:
        frames_total = int(intervention.intervention[FIELDNAME_VIDEO_METADATA][FIELDNAME_FRAMES_TOTAL])

    if frames_total:
        fps = int(intervention.intervention[FIELDNAME_VIDEO_METADATA][FIELDNAME_FPS])

    if not fps:
        warnings.warn("Insufficient Metadata")
    else:
        instance_min_n_frame = instance_min_seconds * fps

        # intervention.update_smooth_labels()

        aggregation = [
            {
                "$match": {FIELDNAME_INTERVENTION_ID: intervention._id}
            },
            {
                "$project": {FIELDNAME_PREDICTIONS_SMOOTH: {"$objectToArray": f"${FIELDNAME_PREDICTIONS_SMOOTH}"}, FIELDNAME_FRAME_NUMBER:1}
            },
            {"$unwind": "$"+FIELDNAME_PREDICTIONS_SMOOTH},
        ]

        images = intervention.db_images.aggregate(aggregation)
        for img in images:
            if img[FIELDNAME_PREDICTIONS_SMOOTH]["v"]:
                _flanks[f"{img[FIELDNAME_PREDICTIONS_SMOOTH]['k']}.{img[FIELDNAME_PREDICTIONS_SMOOTH]['v']}"].append(img[FIELDNAME_FRAME_NUMBER])
        
        for key in _flanks:
            _flanks[key].sort()

        for label_choice, framelist in _flanks.items():
            _start = framelist[0]
            _end = None
            flanks[label_choice][FIELDNAME_FLANK_RANGES] = []
            flanks[label_choice][FIELDNAME_FLANK_FRAMES] = []

            for i, n_frame in enumerate(framelist):
                n_previous_frame = framelist[i-1]

                _new_instance = (n_frame-n_previous_frame) > 1
                if _new_instance:
                    _end = n_previous_frame
                    n_frames_flank = _end - _start
                    if n_frames_flank > instance_min_n_frame:
                        flanks[label_choice][FIELDNAME_FLANK_RANGES].append((_start, _end))
                        _start = n_frame
                        _end = None
                    else:
                        _start = n_frame
                        _end = None
                    
            for range_tuple in flanks[label_choice][FIELDNAME_FLANK_RANGES]:
                flanks[label_choice][FIELDNAME_FLANK_FRAMES].append(
                    [
                        intervention.intervention[FIELDNAME_FRAMES][str(_)] 
                        for _ in range(range_tuple[0], range_tuple[1]+1)
                    ]
                )
        
        intervention.db_interventions.update_one({"_id": intervention._id}, {"$set": {FIELDNAME_PREDICTION_FLANKS: dict(flanks)}})
        intervention.refresh()
        return flanks

def split_df_to_consecutive_frame_dfs(df):
    last_n = None
    groups = []
    group = []

    for index, row in df.iterrows():
        if not last_n:
            last_n = row[FIELDNAME_FRAME_NUMBER]
            group.append(index)
            continue   

        frame_diff = row[FIELDNAME_FRAME_NUMBER] - last_n
        if frame_diff>1:
            groups.append(df.loc[group])
            group = []
        
        group.append(index)
        last_n = row[FIELDNAME_FRAME_NUMBER]

    groups.append(df.loc[group])
    return groups



def dict_of_lists_to_text(_dict) -> str:
    text = ""

    for key, value in _dict.items():
        if isinstance(value, list):
            new = "\n".join(value)
        new = key + "\n" + new
        text = text + "\n" + new

    return text


def generate_token_value_query(report_tokens, patho_tokens, _or: bool = True):
    if _or: condition = "$or"
    else: condition = "$and"

    query = {
        condition: {
            f"{FIELDNAME_TOKENS}.\
                {FIELDNAME_TOKENS_REPORT}.\
                {FIELDNAME_TOKENS_VALUE}": {"$in": report_tokens},
            f"{FIELDNAME_TOKENS}.\
                {FIELDNAME_TOKENS_PATHO}.\
                {FIELDNAME_TOKENS_VALUE}": {"$in": patho_tokens},
        }
    }

    return query

def get_train_image_query(label, value, test_intervention_ids, extend_conditions, extend_agg):
    conditions = [
        field_value_query(f"{FIELDNAME_LABELS}.{label}", value),
        fieldvalue_nin_list_query(FIELDNAME_INTERVENTION_ID, test_intervention_ids)
    ]

    if extend_conditions:
        conditions.extend(extend_conditions)

    aggregation = [match_logical_and_aggregation(conditions)]

    if extend_agg:
        aggregation.extend(extend_agg)

    return aggregation


def check_extern_video_annotation_timestamp(video_key, str_timestamp, db_interventions, time_format = "%Y-%m-%dT%H:%M:%S.%fZ"):
    timestamp = dt.strptime(str_timestamp, time_format)
    intervention = db_interventions.find_one({FIELDNAME_VIDEO_KEY: video_key}, {FIELDNAME_VIDEO_ANNOTATION_TIMESTAMP: 1, FIELDNAME_VIDEO_KEY: 1})

    if FIELDNAME_VIDEO_ANNOTATION_TIMESTAMP in intervention:
        if intervention[FIELDNAME_VIDEO_ANNOTATION_TIMESTAMP]<timestamp:
            return False
        else: 
            return True
    else: 
        return True

def generate_db_extern_video_annotation_dict(annotations, time_format):
    video_annotations = {}

    for annotation in annotations:
        annotation_group_ids = [_["annotationGroupId"] for _ in annotation["sequences"]]
        annotation_group_ids = list(set(annotation_group_ids))
        user_id = annotation["userId"]
        session_date = dt.strptime(annotation["sessionDate"], time_format)
        label_group_id = annotation["labelGroupId"]
        
        for sequence in annotation["sequences"]:
            label = sequence["annotationName"].lower()
            value = sequence["annotationValue"]
            if isinstance(value, str):
                if "true" in value:
                    value = True
                elif "false" in value:
                    value = False
            flank = (sequence["startFrame"],sequence["endFrame"])
            

            if label not in video_annotations:
                video_annotations[label] = {}
            if user_id not in video_annotations[label]:
                video_annotations[label][user_id] = {}
            if session_date not in video_annotations[label][user_id]:
                video_annotations[label][user_id][session_date] = []

            video_annotations[label][user_id][session_date].append(
                {
                    "label": label,
                    "value": value,
                    "flank": flank,
                    "label_group_id": label_group_id
                }
            )

    

    return video_annotations


def select_relevant_extern_video_annotations(video_annotations: dict, annotation_priority_user_id: List):
    selected_annotations = []
    
    for label, user_annotations in video_annotations.items():
        for user_id in annotation_priority_user_id:
            if user_id in user_annotations:
                dates = [_ for _ in user_annotations[user_id].keys()]
                dates.sort(reverse = True)
                selected_annotations.extend(user_annotations[user_id][dates[0]])
    
    return selected_annotations


def add_false_labels_for_test_data_annotations(intervention, extern_video_annotations, annotation_group_labels):
    n_frames_total = int(intervention[FIELDNAME_VIDEO_METADATA][FIELDNAME_FRAMES_TOTAL])
    extern_video_annotations = extern_video_annotations.copy()

    flanks = {_:defaultdict(list) for _ in annotation_group_labels}

    for annotation in extern_video_annotations:
        flanks[annotation["label"]][annotation["value"]].append(annotation["flank"])
        flanks[annotation["label"]]["set_false"] = []
        flanks[annotation["label"]]["not_false"] = []

    for label in flanks.keys():
        for value in flanks[label].keys():
            if value == False or value == "set_false" or value=="not_false":
                continue
            flanks[label]["not_false"].extend(flanks[label][value])


        flanks[label]["not_false"] = list(set(flanks[label]["not_false"]))
        flanks[label]["not_false"].sort(key=itemgetter(0))
        last_end = None
        for sequence in flanks[label]["not_false"]:
            current_start = sequence[0]

            if not last_end:
                if current_start != 0:
                    last_end = 0
                else: 
                    last_end = sequence[1]
                    continue
       
            flanks[label]["set_false"].append((last_end, current_start))
            last_end = sequence[1]
        
        if not last_end:
            flanks[label]["set_false"].append((0, n_frames_total))

        elif last_end < n_frames_total:
            flanks[label]["set_false"].append((last_end, n_frames_total))
            
        elif last_end > n_frames_total:
            last_flank = flanks[label]["set_false"][-1]
            flanks[label]["set_false"][-1] = (last_flank[0], n_frames_total)


        extern_video_annotations.extend([
            {"label": label, "value": False, "flank": flank} for flank in flanks[label]["set_false"]
        ])

        _compare = []
        _unique = []
        for _ in extern_video_annotations:
            to_compare = (_["label"], _["flank"])
            if to_compare in _compare:
                continue
            else: 
                _unique.append(_)
                _compare.append(to_compare)

    return _unique
        


from bson import ObjectId
from pydantic import (
    BaseModel, 
    validator
)
from typing import Union

class PyObjectId(ObjectId):
    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v):
        if not ObjectId.is_valid(v):
            raise ValueError("Invalid objectid")
        return ObjectId(v)

    # @classmethod
    # def __modify_schema__(cls, field_schema):
    #     field_schema.update(type="string")

class Flank(BaseModel):
    name: str
    value: Union[bool, str]
    start: int
    stop: int

    @validator("name")
    def name_to_lower(cls, v):
        return v.lower()


class SettingsSmoothing(BaseModel):
    pass



from pydantic import (
    BaseModel,
    NonNegativeInt
)
from typing import List
from datetime import datetime

from .base import Flank, SettingsSmoothing

class Prediction(BaseModel):
    name: str
    version: float
    date: datetime

class BinaryPrediction(Prediction):
    value: bool
    raw: float
    choices: List[bool]
    label_type = "binary"

class MultilabelPrediction(Prediction):
    value: NonNegativeInt
    raw: List[float]
    choices: List[str]
    label_type = "multilabel"

    def to_binary_predictions(self):
        binary_predictions = []
        for i, choice in enumerate(self.choices):
            prediction_dict = self.dict(include = {
                "version",
                "date"
            })
            prediction_dict["name"] = choice
            prediction_dict["value"] = self.value == i
            prediction_dict["raw"] = self.raw[i]
            prediction_dict["choices"] = [False, True]
            binary_predictions.append(BinaryPrediction(**prediction_dict))
        return binary_predictions

class MultichoicePrediction(Prediction):
    value: List[NonNegativeInt]
    raw: List[float]
    choices: List[str]
    label_type = "multichoice"


class VideoSegmentationPrediction(Prediction):
    value: List[Flank]
    name: str
    settings_smoothing: SettingsSmoothing
    label_type = "video_segmentation"

